import argparse
from pathlib import Path
import matplotlib.pyplot as plt
from daisypy.optim import (
    animate_result,
    plot_convergence,
    plot_result,
    plot_result_nd
)

def main(run_dir):
    run_dir = Path(run_dir)
    result_file = run_dir / 'logs' / 'result.csv'

    ani = animate_result(result_file)
    plt.show()

    fig, ax = plot_result(result_file)
    plt.show()

    fig, ax = plot_result_nd(result_file, ('K_aquitard', 'Z_aquitard'))
    plt.show()

    fig, ax = plot_convergence(result_file)
    plt.show()

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("run_dir", type=str, help='Path to run dir')
    args = parser.parse_args()
    main(run_dir=args.run_dir)



# pylint: disable=R0801
import multiprocessing
import warnings
import numpy as np
import cma
from cma.fitness_transformations import ScaleCoordinates
from cma.optimization_tools import EvalParallel2
from .problem import ScalarProblemWrapper

class DaisyCMAOptimizer:
    """Daisy optimizer using the CMA-ES method from https://github.com/CMA-ES/pycma

     There are many options for cma. The most important for new users is `maxfevals`, which
     limits the number of function evaluations. By default this is set to 10, which is
     unrealisticly low. This can be overridden by setting it to np.inf to get unlimited
     evaluations or to to a finite number, e.g
       cma_options = { "maxfevals" : np.inf }
       cma_options = { "maxfevals" : 2000 }

     A simple estimate of the time it will take to compute N evaluations can be found in this way
       time_to_run_once = <time to run one simulation with Daisy>
       total_run_time = time_to_run_once * maxfevals / number_of_compute_cores
    """
    def __init__(self, problem, logger, cma_options=None, number_of_processes=None):
        """
        Parameters
        ----------
        problem : DaisyProblem

        cma_options : dict
          Options to pass on to cma. See cma.CMAOptions for details
        """
        self.problem = problem
        self.logger = logger
        if number_of_processes is None:
            self.number_of_processes = multiprocessing.cpu_count()
        else:
            self.number_of_processes = number_of_processes
        lower = []
        upper = []
        x0 = []
        for param in problem.parameters:
            lower.append(param.valid_range[0])
            upper.append(param.valid_range[1])
            x0.append(param.initial_value)
        self.objective = ScaleCoordinates(
            ScalarProblemWrapper(problem), lower=lower, upper=upper, from_lower_upper=(-1,1)
        )

        # Map the initial values to optimization domain
        x0 = self.objective.inverse(x0)

        # Setup options
        if cma_options is None:
            cma_options = {}

        if "maxfevals" not in cma_options:
            warnings.warn("Max function evaluations not set, using 10")
            cma_options["maxfevals"] = 10

        if 'bounds'  in cma_options:
            warnings.warn(
                "'bounds' set in cma_options will be ignored and set to match problem parameters"
            )
        cma_options['bounds'] = [-1, 1]
        self.optimizer = cma.CMAEvolutionStrategy(x0, 1/3, cma_options)

    def optimize(self):
        '''Run the optimizer'''
        # pylint: disable=too-many-locals
        max_attempts_to_get_feasible = 3
        # TODO: Implement logging + checkpointing every n'th step
        total_f_evals = 0
        with EvalParallel2(self.objective, self.number_of_processes) as eval_all:
            step = 0
            while not self.optimizer.stop():
                step += 1
                # Try a couple of times if we dont get at least one non nan value
                for i in range(max_attempts_to_get_feasible):
                    xs = self.optimizer.ask()
                    fvals = np.array(eval_all(xs))
                    total_f_evals += len(fvals)
                    if np.any(np.isfinite(fvals)):
                        break
                    self.logger.warning(
                        step=step,msg=f'All are infeasible at attempt {i}', fvals=fvals
                    )
                for x, fval in zip(xs, fvals):
                    raw_params = {
                        f'param_{p.name}' : value  for p, value in
                        zip(self.problem.parameters, self.objective.transform(x))
                    }
                    standardized_params = {
                        f'param_{p.name}' : value  for p, value in
                        zip(self.problem.parameters, x)
                    }
                    objective_value = { f'metric_{self.problem.objective_fn.name}' : fval }
                    self.logger.result(step=step, tag="raw", **objective_value, **raw_params)
                    self.logger.result(
                        step=step, tag="standardized", **objective_value, **standardized_params
                    )

                failed = np.isnan(fvals)
                if np.all(failed):
                    self.logger.error('All attempts failed. Aborting')
                    break

                self.logger.info(step=step, total_function_evaluations=total_f_evals)
                self.logger.info(step=step, median_objective=np.median(fvals[~failed]))
                num_failures = failed.sum()
                if num_failures > 0:
                    self.logger.warning(step=step, n_failed_runs=num_failures)
                    # cma sets nans to the median.
                    # We want them to have a bigger negative influence
                    # TODO: This assumes that are we minimizing ...
                    fvals[failed] = 2*np.max(fvals[~failed])
                self.optimizer.tell(xs, fvals)

                # Log parameter distributions in the standardized space
                means = self.optimizer.result[5]
                stds = self.optimizer.result[6]
                p_mean = {
                    f'param_{p.name}_mean' : mean for p, mean in zip(self.problem.parameters, means)
                }
                p_std = {
                    f'param_{p.name}_std' : std for p, std in zip(self.problem.parameters, stds)
                }
                self.logger.parameters(
                    distribution="normal",
                    tag="standardized",
                    step=step,
                    **p_mean,
                    **p_std
                )

                # Log parameter distributions in the raw space
                means = self.objective.transform(means)
                stds = np.array(self.objective.multiplier) * stds
                p_mean = {
                    f'param_{p.name}_mean' : mean for p, mean in zip(self.problem.parameters, means)
                }
                p_std = {
                    f'param_{p.name}_std' : std for p, std in zip(self.problem.parameters, stds)
                }
                self.logger.parameters(
                    distribution="normal",
                    tag="raw",
                    step=step,
                    **p_mean,
                    **p_std
                )

        status = self.optimizer.result[7]
        self.logger.info('Termination conditions')
        for k, v in status.items():
            self.logger.info(f'{k} = {v}')
        best = self.objective.transform(self.optimizer.result[0])
        means, stds = self.optimizer.result[5], self.optimizer.result[6]
        transformed = self.objective.transform(means)
        result = {
            p.name : {
                'best' : best[i],
                'mean_transformed' : transformed[i],
                'initial_value' : p.initial_value,
                'valid_range' : p.valid_range,
                'mean' : means[i],
                'std' : stds[i],
                'transform' : (self.objective.zero[i], self.objective.multiplier[i])
            } for i, p in enumerate(self.problem.parameters)
        }
        return result

    def checkpoint(self, path):
        '''Save the state to disk to we can resume

        Parameters
        ----------
        path : str
          Path to store checkpoint in
        '''
        # TODO: Save state to disk
        raise NotImplementedError("Checkpointing is not yet implemented")

    @staticmethod
    def from_checkpoint(path):
        '''Read state from disk to we can resume

        Parameters
        ----------
        path : str
          Path to read checkpoint from
        '''
        # TODO: Read the state from disk
        raise NotImplementedError("Resuming from checkpoint is not yet implemented")



'''Functions that formats python objects as string'''

def quote_if_string(value):
    '''Format a value as a string. Add quotes if value is already a string'''
    if isinstance(value, str):
        return '"' + value + '"'
    return str(value)



# pylint: disable=missing-function-docstring
from pathlib import Path
import pandas as pd
from daisypy.optim import ScalarObjective
from daisypy.optim.loss_fns import mse
from .mockup import MockDataExtractor

def test_csv_delimiter():
    in_dir = Path(__file__).parent / 'test-data' / 'targets'
    expected = pd.read_csv(in_dir / 'comma-separated.csv').rename(columns={"NO3" : "value"})
    expected["time"] = pd.to_datetime(expected["time"])
    data_extractor = MockDataExtractor(expected)
    for target_file in in_dir.iterdir():
        if target_file.name.endswith('separated.csv'):
            # We use the same extracted data, but change the target each time.
            f = ScalarObjective(target_file.name, data_extractor, target_file, "NO3", mse)
            result = f(in_dir).pop(target_file.name)
            assert result == 0, target_file



# pylint: disable=R0801
from pathlib import Path
from daisypy.optim import (
    PyFileGenerator, DaiFileGenerator, MultiFileGenerator, DaisyRunner
)
from .markers import requires_daisy

EXPECTED_ONE_FILE = "f(4) = 8, f(-1) = 6.75"
@requires_daisy
def test_run_with_python(tmp_path):
    '''Test that we can run Daisy with generated dai and python files'''
    template_dir = Path(__file__).parent / 'templates'
    generator = MultiFileGenerator({
        'py' : PyFileGenerator('testing.py', template_file_path=template_dir / 'template.py'),
        'dai' : DaiFileGenerator('run.dai', template_file_path=template_dir / 'template.dai')
    })

    params = {
        'py' : { 'a' : 0.25, 'b' : 7 },
        'dai' : { 'x1' : 4, 'x2' : -1 }
    }

    # Assume we are on linux and daisy is installed
    runner = DaisyRunner('daisy')
    dai_file = generator(tmp_path, params)['dai']
    result = runner(dai_file, str(tmp_path))
    assert result.returncode == 0

    daisy_log = Path(tmp_path / 'daisy.log')
    with daisy_log.open(encoding='utf-8') as f:
        lines = list(f)
    assert len(lines) >= 2
    assert lines[-2].strip() == EXPECTED_ONE_FILE


EXPECTED_SEVERAL_FILES = "f(4) = 6, f(-1) = 7.25"
@requires_daisy
def test_run_with_several_python_files(tmp_path):
    '''Test that we can run Daisy with generated dai and multiple python files'''
    template_dir = Path(__file__).parent / 'templates'
    python_dir = Path(__file__).parent / 'py-files'
    # 'py' generates the python module that is called from Daisy. Filename and function name must
    # match what is defined in the 'dai' file.
    # 'util' generates a python script that is imported by 'py'. Filename and function name must
    # match what is defined in the 'py' file.
    # 'dai' generates the file run by Daisy. You should only generate one dai file.
    generator = MultiFileGenerator({
        'py' : PyFileGenerator('testing.py', template_file_path=template_dir / 'template2.py'),
        'util' : PyFileGenerator('util.py', template_file_path=python_dir / 'util.py'),
        'dai' : DaiFileGenerator('run.dai', template_file_path=template_dir / 'template.dai')
    })

    params = {
        'py' : { 'a' : 0.25, 'b' : 7 },
        'dai' : { 'x1' : 4, 'x2' : -1 },
        'util' : {} # No parameters
    }

    # Assume we are on linux and daisy is installed
    runner = DaisyRunner('daisy')
    dai_file = generator(tmp_path, params)['dai']
    result = runner(dai_file, str(tmp_path))
    assert result.returncode == 0

    daisy_log = Path(tmp_path / 'daisy.log')
    with daisy_log.open(encoding='utf-8') as f:
        lines = list(f)
    assert len(lines) >= 2
    assert lines[-2].strip() == EXPECTED_SEVERAL_FILES



import pandas as pd
import numpy as np

def generate_time_series(start_date, end_date, steps, columns):
    '''
    Generate a time series. Data will be either stationary, cyclic or increasing with standard
    normal noise
    '''
    rng = np.random.default_rng()
    data = {
        'time' : pd.date_range(start_date, end_date, steps)
    }

    for col in columns:
        t = rng.random()
        if t < 1/3:
            # Stationary
            data[col] = rng.normal(size=steps)
        elif t < 2/3:
            # cyclic
            data[col] = 2 * np.sin(np.linspace(0, 4*np.pi, steps)) +  rng.normal(size=steps)
        else:
            # Increasing
            data[col] = np.linspace(0, 4, steps) + rng.normal(size=steps)
    return pd.DataFrame(data)



from collections.abc import Sequence
from .util import flatten
from .multi_objective import MultiObjective

class AggregateObjective(Sequence):
    '''Objective that computes several objectives and optionally aggregates them.'''

    def __init__(self, name, objective_fns, aggregate_fn):
        '''
        Parameters
        ----------
        name : str
          Name of objective

        objective_fns : dict of [str, Callable[[str], float]
          Mapping from objective names to objective functions. The objective function is passed the
          path to a daisy output directory and is expected to return a scalar

        aggregate_fn : Callable[[dict of [str, float]], float]
          Function that aggregates the computed objectives. It should map a dict of named objective
          values to a single scalar.
        '''
        self.name = name
        self.multi_objective = MultiObjective('to-be-aggregated', objective_fns)
        self.aggregate_fn = aggregate_fn

    def __call__(self, daisy_output_directory):
        '''Compute the objective

        Parameters
        ----------
        daisy_output_directory : str
          Path to daisy ouput directory that is used when calling the objective functions

        Returns
        -------
        objective_map : dict of [str, float]
          Map from the objective name to the aggregated objective value
        '''
        return { self.name : self.aggregate_fn(self.multi_objective(daisy_output_directory)) }

    def __getitem__(self, index):
        # We could consider flattening objective_fns, but not sure that there is a user case
        # If we do, we also need to update __len__
        return self.multi_objective[index]

    def __len__(self):
        return len(self.multi_objective)

    @property
    def objective_fns(self):
        '''The objective functions being aggregated'''
        return self.multi_objective.objective_fns

    @property
    def variable_name(self):
        '''Names of variables used in aggregated objectives

        Returns
        -------
        list of str
        '''
        return flatten(self.multi_objective, lambda x : x.variable_name)

    @property
    def target(self):
        '''Targets used on aggregated objectives

        Returns
        -------
        list of pandas.DataFrame
        '''
        return flatten(self.multi_objective, lambda x : x.target)

    @property
    def log_name(self):
        '''Names of Daisy log files used in aggregated objectives

        Returns
        -------
        list of str
        '''
        return flatten(self.multi_objective, lambda x : x.log_name)



"""Example showing how to do parameter optimization in Daisy.
The optimization consists of four steps
  1. Setup
  2. Optimize
  3. Evaluate
  4. Analyze
"""
# pylint: disable=missing-function-docstring
import argparse
import os
import json
import warnings
with warnings.catch_warnings():
    # We dont want the Pyarrow depedency warning
    warnings.simplefilter("ignore", category=DeprecationWarning)
    import pandas as pd
import matplotlib.pyplot as plt
from daisypy.io.dlf import read_dlf
from daisypy.optim import (
    DaiFileGenerator,
    available_aggregate_fns,
    DefaultLogger,
    available_loss_fns,
    available_optimizers,
    AggregateObjective,
    ScalarObjective,
    DaisyOptimizationProblem,
    ContinuousParameter,
    DaisyRunner,
)

def main(debug=False):
    daisy_path = "{daisy_path}"
    daisy_home = "{daisy_home}"

    out_dir = "{out_dir}"
    os.makedirs(out_dir, exist_ok=True)

    optimizer_name = "{optimizer}"
    dai_template = "{dai_template}"
    parameters = read_parameters("{parameter_file}")

    # Objective
    log_names = {log_names}
    variable_names = {variable_names}
    target_files = [pd.read_csv(target_file) for target_file in {target_files}]
    loss_fns = [available_loss_fns[loss_fn] for loss_fn in {loss_fns}]
    aggregate_fn = available_aggregate_fns["{aggregate_fn}"]

    objective = setup_objective(log_names, variable_names, target_files, loss_fns, aggregate_fn)

    run_id = get_run_id("{out_dir}")
    run_out_dir = os.path.join(out_dir, run_id)
    log_dir = os.path.join(run_out_dir, 'logs')
    logger = DefaultLogger(log_dir)

    problem = setup(
        daisy_path, daisy_home, run_out_dir, dai_template, parameters, objective, debug
    )
    result = optimize(problem, optimizer_name, logger, run_out_dir)
    eval_dir = evaluate(result, problem, run_out_dir)
    analyze(eval_dir, problem, run_out_dir)

def read_parameters(path):
    with open(path, 'r', encoding='utf-8') as infile:
        parameter_specs = json.load(infile)
    parameters = []
    for spec in parameter_specs:
        assert "type" in spec, 'Missing "type" key'
        assert "name" in spec, 'Missing "name" key'
        assert "initial_value" in spec, 'Missing "initial_value" key'
        assert "valid_range" in spec, 'Missing "valid_range" key'
        if spec["type"] == "continuous":
            parameters.append(
                ContinuousParameter(
                    name=spec["name"],
                    initial_value=spec["initial_value"],
                    valid_range=(spec["valid_range"][0], spec["valid_range"][1])
                )
            )
        else:
            raise ValueError(f'Parameter type: "{{spec["type"]}}" not supported.')
    return parameters

def setup_objective(log_names, variable_names, target_files, loss_fns, aggregate_fn):
    if not len(variable_names) == len(log_names) == len(target_files) == len(loss_fns):
        raise ValueError("Objective specification mismatch")

    # Now we can create the objective functions
    objective_fns = [
        ScalarObjective(log, var, target, loss) for log, var, target, loss in
        zip(log_names, variable_names, target_files, loss_fns)
    ]

    # If we have more than one objective, we wrap in an AggregatorObjective
    if len(objective_fns) == 1:
        return objective_fns[0]
    return AggregateObjective(objective_fns, aggregate_fn)

def setup(daisy_path, daisy_home, base_out_dir, dai_template, parameters, objective, debug):
    # Setup a runner that knows how to execute daisy
    runner = DaisyRunner(daisy_path, daisy_home)

    # Setup a generator that can generate a valid .dai file from a template and parameter values
    file_generator = DaiFileGenerator(template_file_path=dai_template)

    # Create a dirctory for storing setup files for debug/inspection
    out_dir = os.path.join(base_out_dir, 'setup')
    os.makedirs(out_dir, exist_ok=True)

    # Generate a dai file for debug runs
    params = {{ p.name : p.initial_value  for p in parameters }}
    file_generator(out_dir, params, tagged=False)

    # Now we can wrap everything as a DaisyOptimizationProblem that knows how to
    # run daisy, get the output data and compute the objetive
    data_dir = os.path.join(base_out_dir, "debug") if debug else None
    problem = DaisyOptimizationProblem(runner, file_generator, objective, parameters, data_dir, debug)
    return problem

def optimize(problem, optimizer_name, logger, base_out_dir):
    out_dir = os.path.join(base_out_dir, "optimize")
    os.makedirs(out_dir, exist_ok=True)

    # Define some default parameters for each of the optimizers
    optimizer_options = {{
        "sequential" : {{
            "num_samples" : 20, # Number of samples from each continuous parameter
        }},
        "cma" : {{
            "maxfevals" : 100, # Maximum number of function evaluations
        }}
    }}

    options = optimizer_options[optimizer_name]
    optimizer = available_optimizers[optimizer_name](problem, logger, options)

    # Optimize
    result = optimizer.optimize()

    # Print the result
    for name, res in result.items():
        print(name)
        for k,v in res.items():
            print('    ', k, ' : ', v, sep='')

    # Save the result
    with open(os.path.join(out_dir, 'result.json'), 'w', encoding='utf-8') as out:
        json.dump(result, out)

    return result

def evaluate(result, problem, base_out_dir):
    # Create a directory for outputs
    out_dir = os.path.join(base_out_dir, "evaluate")
    os.makedirs(out_dir, exist_ok=True)

    # Generate a dai file with the best parameters
    file_generator = problem.file_generator
    parameters = {{ k : v['best'] for k,v in result.items() }}
    dai_path = file_generator(out_dir, {{'dai' : parameters }})['dai']

    # Run daisy with the newly generated dai file
    problem.runner(dai_path, out_dir)
    return out_dir

def analyze(eval_dir, problem, base_out_dir):
    out_dir = os.path.join(base_out_dir, 'analyze')
    os.makedirs(out_dir, exist_ok=True)

    # Extract variables, log names and targets from the problem
    variables = problem.objective_fn.variable_name
    targets = problem.objective_fn.target
    log_names = problem.objective_fn.log_name

    # Check if we only have one triple to process
    if isinstance(variables, str) and \
       isinstance(targets, pd.DataFrame) and \
       isinstance(log_names, str):
        variables = [variables]
        targets = [targets]
        log_names = [log_names]

    # Plot all variables we optimized for.
    # We can have multiple objectives defined for the same (variable, log) pair, in which case we
    # only plot it once.
    processed = set()
    for i, (var, target, log_name) in enumerate(zip(variables, targets, log_names)):
        if (var, log_name) in processed:
            continue
        processed.add((var, log_name))
        plot_single_variable(var, target, log_name, eval_dir, out_dir)

def plot_single_variable(var, target, log_name, eval_dir, out_dir):
    # Get the simulated results
    sim_dlf = read_dlf(os.path.join(eval_dir, log_name))
    sim_value = sim_dlf.body[var]
    sim_time = pd.to_datetime(
        sim_dlf.body[['year', 'month', 'mday', 'hour']].rename(columns={{'mday' : 'day'}})
    )
    sim = pd.DataFrame({{'time' : sim_time, 'value' : sim_value}})

    log_display_name = os.path.splitext(log_name)[0]
    title = f"{{var}} @ {{log_display_name}}"
    prefix = f"{{var}}-{{log_display_name}}"

    # Make a time series plot of the target and simulated values
    fig, ax = plt.subplots()
    ax.plot(target["time"], target['value'])
    ax.plot(sim["time"], sim['value'], linestyle='dashed')
    ax.legend(["target", "simulation"])
    ax.set_ylabel(var)
    ax.set_title(title)
    fig.tight_layout()
    fig.savefig(os.path.join(out_dir, f'{{prefix}}-time-series.pdf'))

    # Make a scatter plot of the tarvet vs simulated value
    fig, ax = plt.subplots()
    merged = pd.merge(
        target, sim,
        how='left',
        on='time',
        suffixes=('_target', '_sim')
    )
    ax.scatter(merged['value_target'], merged['value_sim'])
    ax.axis('equal')
    ax.axline((merged['value_target'][0],)*2, slope=1, color='red')
    ax.set_xlabel('Target')
    ax.set_ylabel('Simulated')
    ax.set_title(title)
    fig.tight_layout()
    fig.savefig(os.path.join(out_dir, f'{{prefix}}-scatter.pdf'))
    plt.show()

def get_run_id(base_out_dir):
    used = set()
    for entry in os.scandir(base_out_dir):
        try:
            used.add(int(entry.name))
        except ValueError:
            pass
    runid = 0
    try:
        while True:
            used.remove(runid)
            runid += 1
    except KeyError:
        return str(runid)

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--debug", action="store_true")
    args = parser.parse_args()
    main(debug=args.debug)



import os
from .file_generator import FileGenerator

class PyFileGenerator(FileGenerator):
    """Template based generation of python files using string replacement

    Parameters in the template are specifed in curly braces {}. For example,

        active_depth = {active_soil_layer_depth}

    Which specifies a parameter called `active_soil_layer_depth`.

    Note:
    In the template, you need to use double braces in sets and f-strings.

    my_set = {{ my_var }}
    my_string = f'{{ my_var }}
    """
    def __init__(self, out_file, template_text='', template_file_path=None):
        """
        Parameters
        ----------
        out_file : str
          Name to use for generated file. This needs to match what you use in the dai file

        template_text : str
          Template text.

        template_file_path : str
          Path to template. Overrides template_text if not None
        """
        self.out_file = out_file
        if template_file_path is not None:
            with open(template_file_path, 'r', encoding='utf-8') as infile:
                # Skip python line comments
                self.template_text = ''.join((
                    line for line in infile if not line.lstrip().startswith('#')
                ))
        else:
            self.template_text = template_text

    def __call__(self, output_directory, params, tagged=True):
        """Generate a python file from the template using the given params and write it to a
        directory

        Parameters
        ----------
        output_directory : str
          Directory to store the generated file in

        params : dict (str, value) OR { 'py' : dict (str, value) }
          If tagged is True, then the key 'py' MUST be in params and the value MUST be a dict of
          parameters, where the keys MUST match the defined template parameters exactly.
          If tagged is False, then the keys MUST match the defined template parameters exactly.

        tagged : bool
          If True return a tagged path otherwise return a plain path

        Returns
        -------
        { 'py' : out_path } OR out_path
        """
        if tagged:
            params = params['py']
        os.makedirs(output_directory, exist_ok=True)
        py_string = self.template_text.format(**params)
        out_path = os.path.abspath(os.path.join(output_directory, self.out_file))
        with open(out_path, "w", encoding='utf-8') as f:
            f.write(py_string)
        if tagged:
            return { 'py' : out_path }
        return out_path

    def serialize(self):
        '''Serializable representation of this PyFileGenerator

        Returns
        -------
        dict of (str, str)
        '''
        return {
            'template_text' : self.template_text,
            'out_file' : self.out_file
        }

    @staticmethod
    def unzerialize(dict_repr):
        '''Create a PyFileGenerator from a serialized representation

        Parameters
        ----------
        dict_repr: dict of (str, str)
          dict with keys 'template_text' and 'out_file'
        '''
        return PyFileGenerator(template_text=dict_repr['template_text'],
                                out_file=dict_repr['out_file'])



from daisypy.optim.terminal_log import TerminalLog

def test_terminal_log(capsys):
    '''Test that TerminalLog prints to the expected streams (stdout, stderr)'''
    expected_out = '\n'.join([
        'tag=test,step=1,value=0.1,message=string',
        'test,2,0.2,string 2',
    ])
    expected_err = '\n'.join([
        'done,step=3,value=0.3',
    ])
    with TerminalLog() as info, TerminalLog(error=True) as error:
        info.log(tag='test', step=1, value=0.1, message="string")
        info.log('test', 2, 0.2, "string 2")
        error.log('done', step=3, value=0.3)

    captured = capsys.readouterr()
    assert captured.out.strip() == expected_out
    assert captured.err.strip() == expected_err



import subprocess
import pytest

def has_working_daisy():
    '''Check that daisy can be called and "works"'''
    # pylint: disable=broad-exception-caught
    try:
        return subprocess.run(["daisy", "-v"], check=False).returncode == 0
    except Exception:
        return False

requires_daisy = pytest.mark.skipif(
    not has_working_daisy(),
    reason="No working daisy installation found"
)



"Setup input files and script for running a Daisy optimization"
import argparse
import os
import sys
import json
import csv
import importlib.resources
import platform
import subprocess
import shutil

from daisypy.optim import (
    available_aggregate_fns,
    available_loss_fns,
    available_optimizers,
)

def main():
    '''Entry point for create'''
    # We need argument parsing in the entry point function
    parser = argparse.ArgumentParser(
        description=__doc__
    )
    parser.add_argument("--example", action="store_true")
    args = parser.parse_args()
    # pylint: disable=broad-exception-caught
    # Regardless of what happens we want to try and inform the user
    try:
        run(args.example)
        sys.exit(0)
    except Exception as e:
        print("Error creating project", e)
        sys.exit(1)

def run(example=False):
    '''
    Parameters
    ----------
    example : bool
      If True generate example optimization, otherwise prompt user for inputs
    '''
    if example:
        config = get_example_config()
    else:
        config = get_config()
    create_optim(config)
    finalize()

def get_example_config():
    '''Generate example configuration. User is prompted for Daisy home and Daisy executable paths'''
    config = {
        'name': 'example',
        'out_dir': 'out',
        'optimizer': 'sequential',
        'dai_template': 'input/template.dai',
        'parameter_file': 'input/parameters.json',
        'num_continuous_parameters': 0,
        'objectives' : {
            0 : {
                'variable_name': 'N2O-Nitrification',
                'log_name': 'field_nitrogen.dlf',
                'target_file': 'input/target.csv',
                'loss_fn': 'mse',
            },
            1 : {
                'variable_name': 'N2O-Nitrification',
                'log_name': 'field_nitrogen.dlf',
                'target_file': 'input/target.csv',
                'loss_fn': 'mae'
            }
        },
        'aggregate_fn' : 'sum'
    }
    config["daisy_home"] = get_daisy_home()
    config["daisy_path"] = get_daisy_path(config['daisy_home'])
    return config

def get_config():
    '''Prompt user for full configuration'''
    config = {}
    config["name"] = input_with_default("Name", "my-optimization", lambda x: not os.path.exists(x))
    config["daisy_home"] = get_daisy_home()
    config["daisy_path"] = get_daisy_path(config["daisy_home"])
    config["out_dir"] = input_with_default("Output directory", "out")

    config["optimizer"] = input_with_choices("Optimization method",
                                             list(available_optimizers.keys()),
                                             None,
                                             False)
    ##  TODO: Get optimizer specific parameters
    config["dai_template"] = input_with_default("Dai template file", "input/template.dai")
    config["parameter_file"] = input_with_default("Parameter file", "input/parameters.json")
    if not os.path.exists(config["parameter_file"]):
        config["num_continuous_parameters"] = int(
            input_with_default("Number of continuous parameters", 0, is_int)
        )
        # config["num_categorical_parameters"] = int(
        #     input_with_default("Number of categorical parameters", 0, is_int)
        # )
    objectives = {}
    add_more = True
    while add_more:
        objective, add_more = _prompt_for_objective()
        objectives[len(objectives)] = objective
    config["objectives"] = objectives
    if len(objectives) > 1:
        config["aggregate_fn"] = input_with_choices("Aggregate function",
                                                    list(available_aggregate_fns.keys()))
    else:
        config["aggregate_fn"] = "sum"

    return config

def _prompt_for_objective():
    objective = {}
    objective["variable_name"] = input_no_default("Name of variable to optimize for")
    objective["log_name"] = input_no_default("Name of log file where the variable is logged")
    objective["target_file"] = input_with_default("Path to target file", "input/target.csv")
    objective["loss_fn"] = input_with_choices("Loss function", list(available_loss_fns.keys()))
    add_more = 'y' == input_no_default("Add more objectives? (y/n)",
                                       check=lambda s: s in ('y', 'n'))
    return objective, add_more

def create_optim(config):
    '''Create an optimization project from a configuration'''
    config["daisy_home"] = sanitize_path(os.path.abspath(config["daisy_home"]))
    config["daisy_path"] = sanitize_path(os.path.abspath(config["daisy_path"]))
    basedir = os.path.abspath(config["name"])
    os.makedirs(basedir, exist_ok=False)
    # This is two calls because we want to fail if the basedir already exists
    os.makedirs(os.path.join(basedir, "input"))

    if not os.path.isabs(config["out_dir"]):
        config["out_dir"] = os.path.join(basedir, config["out_dir"])
    config["out_dir"] = sanitize_path(config["out_dir"])

    # Copy or create parameter file
    config["parameter_file"] = _copy_or_create_param(config, basedir)

    # Copy or create target files fot objectives
    # And transform from dict of objectives to lists
    objectives = config.pop("objectives")
    for objective_id, objective in objectives.items():
        # Copy or create target file
        objective["target_file"] = _copy_or_create_target(objective_id, objective, basedir)

    config['variable_names'] = [ obj["variable_name"] for obj in objectives.values() ]
    config['log_names'] = [ obj["log_name"] for obj in objectives.values() ]
    config['target_files'] = [ obj["target_file"] for obj in objectives.values() ]
    config['loss_fns'] = [ obj["loss_fn"] for obj in objectives.values() ]

    # Copy or create template file
    config["dai_template"] = _copy_or_create_template(config, basedir)

    # Instantiate an optimize.py file
    optimize_template = _read_optimize_template()
    optimize_program = optimize_template.format(**config)
    optimize_path = os.path.join(basedir, "optimize.py")
    with open(optimize_path, "w", encoding='utf-8') as outfile:
        outfile.write(optimize_program)

    # Copy the analyze script
    _copy_script('analyze.py', basedir)

def _copy_or_create_param(config, basedir):
    param_path = os.path.join(basedir, "input", "parameters.json")
    if os.path.exists(config["parameter_file"]):
        try:
            shutil.copyfile(config["parameter_file"], param_path)
        except shutil.SameFileError:
            pass # This is not a problem
    else:
        if config["num_continuous_parameters"] == 0:
            params = get_example_params()
        else:
            params = [{
                "type" : "continuous",
                "name" : f"p{i}",
                "initial_value" : 0,
                "valid_range" : (0, 1)
                } for i in range(config["num_continuous_parameters"])
            ]
        with open(param_path, 'w', encoding='utf-8') as outfile:
            json.dump(params, outfile, indent="  ")
    return sanitize_path(param_path)

def _copy_or_create_target(objective_id, objective, basedir):
    target_path = os.path.join(basedir, "input", f"target-{objective_id}.csv")
    if os.path.exists(objective["target_file"]):
        try:
            shutil.copyfile(objective["target_file"], target_path)
        except shutil.SameFileError:
            pass # This is not a problem
    else:
        with open(target_path, 'w', encoding='utf-8') as outfile:
            writer = csv.writer(outfile)
            for row in get_example_target():
                writer.writerow(row)
    return sanitize_path(target_path)

def _copy_or_create_template(config, basedir):
    template_path = os.path.join(basedir, "input", "template.dai")
    if os.path.exists(config["dai_template"]):
        try:
            shutil.copyfile(config["dai_template"], template_path)
        except shutil.SameFileError:
            pass # This is not a problem
    else:
        # We have a default with some instructions
        # From python 3.13 we can do
        # with importlib.resources.path("daisypy.optim",
        #                               "data",
        #                               "default-template.dai") as default_path:
        with importlib.resources.path("daisypy.optim", "data") as datadir:
            shutil.copyfile(datadir / "default-template.dai", template_path)
    return sanitize_path(template_path)

def _read_optimize_template():
    # From python 3.13 we can do
    # optimize_template = importlib.resources.read_text("daisypy.optim",
    #                                                   "data",
    #                                                   "optimize_template.py",
    #                                                   encoding="utf-8")
    with importlib.resources.path("daisypy.optim", "data") as datadir:
        optimize_inpath = os.path.join(datadir, "optimize_template.py")
    with open(optimize_inpath, 'r', encoding='utf-8') as infile:
        optimize_template = infile.read()
    return optimize_template

def _copy_script(file_name, out_dir):
    out_path = os.path.join(out_dir, file_name)
    with importlib.resources.path("daisypy.optim", "data") as datadir:
        in_path = os.path.join(datadir, file_name)
        try:
            shutil.copyfile(in_path, out_path)
        except (shutil.SameFileError, IOError):
            print(f'Could not copy {file_name}')

def finalize():
    '''Try to add dependencies with uv. Inform user if it fails'''
    cmd = [
        "uv",
        "add",
        "daisypy-optim@git+https://github.com/daisy-model/daisypy-optim",
        "pandas",
        "matplotlib"
    ]
    if 'cma' in available_optimizers:
        cmd.append("cma")
    if 'ax' in available_optimizers:
        cmd.append("ax-platform")

    result = subprocess.run(cmd, text=True, check=False)
    if result.returncode != 0:
        print("Error adding dependencies. Returncode", result.returncode)
        print("stdout:", result.stdout)
        print("stderr:", result.stderr)
        print("Please verify/add manually")
        print(*cmd)

def get_daisy_home():
    '''Prompt user for Daisy home location. Try to guess based on platform'''
    daisy_candidates = []
    match platform.system():
        case "Windows":
            try:
                for entry in os.scandir("C:/Program Files"):
                    if entry.name.startswith("daisy"):
                        daisy_candidates.append(entry.path)
            except FileNotFoundError:
                pass
        case "Linux":
            # Try to get Daisy home from flatpak install
            path = os.path.expanduser('~/.var/app/dk.ku.daisy/data')
            if os.path.exists(path):
                daisy_candidates.append(path)
        case _:
            pass

    match len(daisy_candidates):
        case 0:
            return input_no_default("Path to daisy directory", os.path.isdir)
        case 1:
            return input_with_default("Path to daisy directory", daisy_candidates[0], os.path.isdir)
        case _:
            return input_with_choices("Path to daisy directory", daisy_candidates, os.path.isdir)

def get_daisy_path(daisy_home):
    '''Prompt user for Daisy executable path. Try to guess based on daisy_home/platform'''
    match platform.system():
        case "Windows":
            default = os.path.join(daisy_home, "bin", "daisy.exe")
            if os.path.exists(default):
                return input_with_default("Path to daisy.exe", default, os.path.exists)
            return input_no_default("Path to daisy.exe", os.path.exists)

        case "Linux":
            default = os.path.expanduser("~/.local/bin/daisy")
            if os.path.exists(default):
                return input_with_default("Path to daisy binary", default, os.path.exists)
            return input_no_default("Path to daisy binary", os.path.exists)

        case _:
            default = os.path.join(daisy_home, "bin", "daisy")
            if os.path.exists(default):
                return input_with_default("Path to daisy binary", default, os.path.exists)
            return input_no_default("Path to daisy binary", os.path.exists)


def input_with_default(prompt, default, check=None):
    '''Prompt for user input with a default.
    Optionally validate input.'''
    while True:
        response = input(f"{prompt} [{default}]: ")
        if len(response) == 0:
            print_selected(default)
            return default
        if check is None or check(response):
            print_selected(response)
            return response

def input_no_default(prompt, check=None):
    '''Prompt for user input without a default.
    Optionally validate input.'''
    while True:
        response = input(f"{prompt}: ")
        if check is None or check(response):
            print_selected(response)
            return response

def input_with_choices(prompt, choices, check=None, user_supplied_ok=True):
    '''Prompt for user input with a list of choices.
    Optionally validate input.
    Optinally allow user supply their own choice'''
    while True:
        print(prompt, *[f"  {i} {choice}" for i, choice in enumerate(choices)], sep='\n')
        response = input("Choice: ")
        try:
            i = int(response)
            if 0 <= i < len(choices):
                print_selected(choices[i])
                return choices[i]
        except ValueError:
            if user_supplied_ok and (check is None or check(response)):
                print_selected(response)
                return response

def print_selected(selected):
    '''Print user selection'''
    print(f'  "{selected}"')

def is_int(s):
    '''Check if s can be converted to an integer'''
    try:
        int(s)
        return True
    except ValueError:
        return False

def sanitize_path(path):
    '''Get rid of backslashes in paths'''
    return path.replace("\\", "/")

def get_example_params():
    '''Parameters to optimize in example'''
    return [
        {
            "type" : "continuous",
            "name" : "K_aquitard",
            "initial_value": 1.0,
            "valid_range" : [0.1, 2]
        },
        {
            "type" : "continuous",
            "name" : "Z_aquitard",
            "initial_value" : 150,
            "valid_range" : [50, 250]
        }
    ]

def get_example_target():
    '''The output from running with K_aquitard = 0.456 and Z_aquitard = 236'''
    return [
        ["time", "N2O-Nitrification"],
        ["2000-01-01", "0.000011"],
        ["2000-02-01", "0.029358"],
        ["2000-03-01", "0.077190"],
        ["2000-04-01", "0.312162"],
        ["2000-05-01", "1.034610"],
        ["2000-06-01", "0.490148"],
        ["2000-07-01", "0.214898"],
        ["2000-08-01", "0.217915"],
        ["2000-09-01", "0.180642"],
        ["2000-10-01", "0.163631"],
        ["2000-11-01", "0.280886"],
        ["2000-12-01", "0.204323"],
    ]


if __name__ == '__main__':
    main()



# pylint: disable=too-many-locals,R0801
"""Example showing how to optimize two Daisy parameters for a single objective using CMA"""
import argparse
from pathlib import Path
import pandas as pd
from daisypy.optim import (
    DaiFileGenerator,
    DlfDataExtractor,
    DaisyCMAOptimizer,
    ScalarObjective,
    DaisyOptimizationProblem,
    ContinuousParameter,
    DaisyRunner,
    DefaultLogger
)

# We use the multiprocessing module, which uses pickle, so we cannot use local functions
def ssd(actual, target):
    '''Sum of squared distance'''
    return ((actual - target)**2).sum()

def single_objective_two_parameters_cma(daisy_path, daisy_home):
    '''How to optimize parameters for Daisy

    0. Define a runner that can run Daisy
    1. Setup the dai file generator
    2. Define the parameters that we will optimize
    3. Define the objective
    4. Wrap everything as an optimization problem
    5. Setup a logger
    6. Choose an optimizer
    7. Run the optimizer
    8. Look at the results
    '''
    base_dir = Path(__file__).parent
    out_dir = base_dir / 'out' / 'single-objective-two-parameters-cma'
    data_dir = base_dir / 'example-data'

    # 0. Define a runner that can run Daisy
    runner = DaisyRunner(daisy_path, daisy_home)

    # 1. Setup the dai file generator
    dai_template = data_dir / 'template.dai'
    dai_file_generator = DaiFileGenerator(template_file_path=dai_template)

    # 2. Define the parameters that we will optimize
    # Names of parameters should match the names in the template file
    parameters = [
        ContinuousParameter(
            name='K_aquitard',
            initial_value=0.2,
            valid_range=(0.1, 0.7)
        ),
        ContinuousParameter(
            name='Z_aquitard',
            initial_value=200,
            valid_range=(150, 250)
        ),
    ]

    # 3. Define the objective
    # We need a target, a loss function and the name of the variable and dlf file
    # The target must be a dataframe with a "time" column
    target = pd.read_csv(data_dir / 'measured-field-nitrogen.csv')
    target["time"] = pd.to_datetime(target[['year', 'month', 'day', 'hour']])
    target_name = "NO3-Denitrification"

    # The loss function can be any python function mapping a pair of numpy.arrays to a scalar
    loss_fn = ssd

    # We need a DlfDataExtractor that can extract the variable we are optimizing for
    data_extractor = DlfDataExtractor({
        'field_nitrogen.dlf' : 'NO3-Denitrification'
    })
    objective_fn = ScalarObjective("NO3_Error", data_extractor, target, target_name, loss_fn)

    # 4. Wrap everything as an optimization problem
    # Normally we would not set data_dir and we would set debug = False,
    # but here we set them so we can inspect the output.
    # If debug = False, then outputs are deleted as soon as the optimizer is done with them
    out_data_dir = out_dir / 'data_dir'
    debug = True
    problem = DaisyOptimizationProblem(
        runner, dai_file_generator, objective_fn, parameters, out_data_dir, debug
    )

    # 5. Setup a logger
    # We use DefaultLogger that logs parameter distributions and sampled parameters to csv files
    log_dir = out_dir / 'logs'
    logger = DefaultLogger(log_dir)

    # 6. Setup an optimizer
    # We choose cma. You should try sequential as well.
    # For cma we should always explicitly set the maximum number of function evaluations AKA the
    # maximum number of times we will run Daisy. We set it very low
    cma_options = {
        "maxfevals" : 50
    }
    optimizer = DaisyCMAOptimizer(problem, logger, cma_options)

    # 7. Run the optimizer
    result = optimizer.optimize()

    # 8. Look at the results
    for name, res in result.items():
        print(name)
        for k,v in res.items():
            print('    ', k, ' : ', v, sep='')

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('daisy_path', type=str, help='Path to daisy binary')
    parser.add_argument('--daisy_home', type=str, default=None,
                        help='Path to daisy home directory containing lib/ and sample/')
    args = parser.parse_args()
    single_objective_two_parameters_cma(args.daisy_path, args.daisy_home)



# pylint: disable=too-few-public-methods
import pandas as pd

class LossWrapper:
    """Loss wrappeer for use with DaisyObjective"""
    def __init__(self, loss_fn):
        """
        Parameters
        ----------
        loss_fn : callable : (actual, target) -> loss
          actual and target are numpy.ndarray of shape (n,)
        """
        self.loss_fn = loss_fn

    def __call__(self, actual, target):
        """Compute the loss over a timeseries

        Timestamps MUST be unique in both actual and target.
        All timestamps in target MUST be in actual.
        Extra timestamps in actual are ignored.

        Parameters
        ----------
        actual : pd.DataFrame
          Must contain columns "time" and "value"

        target : pd.DataFrame
          Must contain columns "time" and "value"

        Raises
        ------
        ValueError if timestamps are not unique or if actual is missing timestamps

        Returns
        -------
        loss as computed by self.loss_fn
        """
        # Only compute loss for time point with measurements
        target = target.dropna()
        target_time = target['time']
        if target_time.nunique() != len(target_time):
            raise ValueError('Timestamps in target must be unique')

        actual_time = actual['time']
        if actual_time.nunique() != len(actual_time):
            raise ValueError('Timestamps in actual must be unique')

        if not target_time.isin(actual_time).all():
            raise ValueError('All timestamps in target must be in actual')

        merged = pd.merge(
            target, actual,
            how='left',
            on='time',
            suffixes=('_target', '_actual')
        )
        return self.loss_fn(merged['value_actual'].to_numpy(), merged['value_target'].to_numpy())



# pylint: disable=missing-function-docstring,R0801
from pathlib import Path
from daisypy.optim import (
    PyFileGenerator, DaiFileGenerator, MultiFileGenerator
)

EXPECTED = {
    'py' : """def linear(x):
    return 0.25 * x + 7""",

    'dai' : """(deffunction f Python
  "Call Python function." (module "testing") (name "linear") (domain []) (range []))
(defprogram print_it write
  "Write specific value"
  (declare v1 Number [] "V1")
  (declare v2 Number [] "V2")
  (v1 (apply f 4 []))
  (v2 (apply f -1 []))
  (what "f(4) = ${v1}, f(-1) = ${v2}"))
(run print_it)"""
}

def test_tagged(tmp_path):
    template_dir = Path(__file__).parent / 'templates'
    generator = MultiFileGenerator({
        'py' : PyFileGenerator('testing.py', template_file_path=template_dir / 'template.py'),
        'dai' : DaiFileGenerator('run.dai', template_file_path=template_dir / 'template.dai')
    })

    params = {
        'py' : { 'a' : 0.25, 'b' : 7 },
        'dai' : { 'x1' : 4, 'x2' : -1 }
    }

    paths = generator(tmp_path, params)
    assert len(paths) == len(EXPECTED)
    for gen_name, file_path in paths.items():
        assert Path(file_path).read_text(encoding='utf-8').strip() == EXPECTED[gen_name]

def test_not_tagged(tmp_path):
    template_dir = Path(__file__).parent / 'templates'
    generator = MultiFileGenerator({
        'py' : PyFileGenerator('testing.py', template_file_path=template_dir / 'template.py'),
        'dai' : DaiFileGenerator('run.dai', template_file_path=template_dir / 'template.dai')
    })

    params = {
        'py' : { 'a' : 0.25, 'b' : 7 },
        'dai' : { 'x1' : 4, 'x2' : -1 }
    }

    paths = generator(tmp_path, params, tagged=False)
    assert len(paths) == len(EXPECTED)
    for file_path, expected in zip(paths, EXPECTED.values()):
        assert Path(file_path).read_text(encoding='utf-8').strip() == expected


def test_no_params(tmp_path):
    py_template = 'x = {{ "a" : 2 }}'
    dai_template = '(defprogram print_it write\n  (what "${{v1}}"))'
    expected_all = [ 'x = { "a" : 2 }', '(defprogram print_it write\n  (what "${v1}"))' ]
    generator = MultiFileGenerator({
        'py' : PyFileGenerator('testing.py', template_text=py_template),
        'dai' : DaiFileGenerator('run.dai', template_text=dai_template)
    })

    params = {
        'py' : {},
        'dai' : {}
    }

    paths = generator(tmp_path, params, tagged=False)
    assert len(paths) == len(expected_all)
    for file_path, expected in zip(paths, expected_all):
        assert Path(file_path).read_text(encoding='utf-8').strip() == expected



"""Aggregate functions.
Signature is (objectives : [float]) -> float
"""

def mean(objectives):
    """Mean value of objectives"""
    return sum(objectives) / float(len(objectives))

available_aggregate_fns = {
    "sum" : sum,
    "mean" : mean
}



import os
import warnings
from pathlib import Path
from daisypy.io import parse_dai, format_dai, filter_dai
from daisypy.io.dai import Definition, Comment, Identifier
from .file_generator import FileGenerator

class DaiFileGenerator(FileGenerator):
    """Template based generation of dai files using string replacement

     Parameters in the template are specifed in curly braces {}. For example,

       ...
       (Groundwater aquitard
         (K_aquitard {K_aquitard_param} [mm/d])
         ...
       )

     Which specifies a parameter called `K_aquitard_param`
     """
    def __init__(self, out_file='run.dai', template_text='', template_file_path=None):
        """
        Parameters
        ----------
        out_file : str
          Name to use for generated file

        template_text : str
          Template text.

        template_file_path : str
          Path to template. Overrides template_text if no None

        tag : str
          Tag to use when returning generated paths
        """
        self.out_file = out_file
        if template_file_path is not None:
            template_text = Path(template_file_path).read_text(encoding='utf-8')
        # Parse the text as a Dai object while allowing placeholders
        dai = parse_dai(template_text, extended=True)
        dai = filter_dai(dai, lambda x : not isinstance(x, Comment))

        # Force all programs that inherits from spawn to run with 1 process
        for value in dai.values:
            if isinstance(value, Definition) and value.parent.value == 'spawn':
                has_parallel = False
                for param in value.body:
                    if isinstance(param, list) and param[0].value == 'parallel':
                        has_parallel = True
                        if param[1] != 1:
                            warnings.warn("parallel parameter for spawn forced to 1")
                            param[1] = 1
                if not has_parallel:
                    warnings.warn("parallel parameter for spawn forced to 1")
                    value.body.append([Identifier('parallel'), 1])
        self.template_text = format_dai(dai)

    def __call__(self, output_directory, params, tagged=True):
        """Generate a dai file from the template using the given params and write it to a directory

        Parameters
        ----------
        output_directory : str
          Directory to store the generated file in

        params : dict (str, value) OR { 'dai' : dict (str, value) }
          If tagged is True, then the key 'dai' MUST be in params and the value MUST be a dict of
          parameters, where the keys MUST match the defined template parameters exactly.
          If tagged is False, then the keys MUST match the defined template parameters exactly.

        tagged : bool
          If True return a tagged path otherwise return a plain path

        Returns
        -------
        { 'dai' : out_path } OR out_path
        """
        if tagged:
            params = params['dai']
        os.makedirs(output_directory, exist_ok=True)
        dai_string = self.template_text.format(**params)
        out_path = os.path.abspath(os.path.join(output_directory, self.out_file))
        with open(out_path, "w", encoding='utf-8') as f:
            f.write(dai_string)
        if tagged:
            return { 'dai' : out_path }
        return out_path

    def serialize(self):
        '''Serializable representation of this DaiFileGenerator

        Returns
        -------
        dict of (str, str)
        '''
        return {
            'template_text' : self.template_text,
            'out_file' : self.out_file
        }

    @staticmethod
    def unzerialize(dict_repr):
        '''Create a DaiFileGenerator from a serialized representation

        Parameters
        ----------
        dict_repr: dict of (str, str)
          dict with keys 'template_text' and 'out_file'
        '''
        return DaiFileGenerator(template_text=dict_repr['template_text'],
                                out_file=dict_repr['out_file'])



import os
from daisypy.optim.terminal_log import TerminalLog
from daisypy.optim.csv_log import CsvLog
from daisypy.optim.logger import Logger

available_logs = {
    "terminal" : TerminalLog,
    "csv" : CsvLog
}

def DefaultLogger(outdir): # pylint: disable=invalid-name # (It should look like a class)
    '''Return a logger instance with the following predefined logs.

      'default' : Log to stdout
      'warning', 'error' : Log to stderr
      'parameters' : Log to csv file 'parameters.csv' in outdir
      'result' : Log to csv file 'result.csv' in outdir

    Parameters
    ----------
    outdir : str
      Directory to store log files in

    Returns
    -------
    daisypy.optim.Logger
    '''
    logs = {
        'default' : TerminalLog(),
        'warning' : TerminalLog(error=True),
        'error' : TerminalLog(error=True),
        'parameters' : CsvLog(os.path.join(outdir, 'parameters.csv')),
        'result' : CsvLog(os.path.join(outdir, 'result.csv')),
    }
    return Logger(**logs)



import sys
from .log import Log

class TerminalLog(Log):
    '''A simple terminal log that prints to stdout or stderr'''

    def __init__(self, error=False):
        self.is_open = True
        self.out = sys.stderr if error else sys.stdout

    def log(self, *args, **kwargs):
        if self.is_open:
            args_msg = ','.join([str(a) for a in args])
            kwargs_msg = ','.join([f'{k}={v}' for k,v in kwargs.items()])
            if len(args_msg) > 1 and len(kwargs_msg) > 1:
                msg = args_msg + ',' + kwargs_msg
            else:
                msg = args_msg + kwargs_msg
            print(msg, file=self.out)

    def close(self):
        self.is_open = False

    def __del__(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



# pylint: disable=too-many-locals,R0801
"""Example showing how to optimize parameters for python functions used in Daisy"""
import argparse
from pathlib import Path
from daisypy.optim import (
    ContinuousParameter,
    DefaultLogger,
    DaisyCMAOptimizer,
    DlfDataExtractor,
    PyFileGenerator,
    DaiFileGenerator,
    MultiFileGenerator,
    DaisyRunner,
    DaisyOptimizationProblem,
    ScalarObjective,
    mse
)


def single_objective_with_python(daisy_path, daisy_home=None):
    '''Optimize two parameters in a dai file using a single scalar objective and CMA

    daisy_path: str
      Path to daisy binary

    daisy_home: str
      Path to daisy home. If None let the Daisy binary figure it out
    '''
    base_dir = Path(__file__).parent
    out_dir = base_dir / 'out' / 'single_objective_with_python'
    data_dir = base_dir / 'example-data' / 'python-chemical-reaction'
    file_generator = MultiFileGenerator({
        'py' : PyFileGenerator('daisy-react.py', template_file_path=data_dir / 'daisy-react.py'),
        'dai' : DaiFileGenerator('run.dai', template_file_path=data_dir / 'test-pyreact.dai'),
    })
    data_extractor = DlfDataExtractor({
        'soil_NO3_profile.dlf' : 'NO3'
    })

    parameters = {
        'py' : [ ContinuousParameter('param', 0.24, (0.01, 0.5)) ],
        'dai' : [ ContinuousParameter('clay', 5, (1, 10)) ],
    }

    runner = DaisyRunner(daisy_path, daisy_home)

    target_file = data_dir / 'target.csv'
    objective = ScalarObjective(
        name="NO3",
        data_extractor=data_extractor,
        target=target_file,
        target_name="NO3",
        loss_fn=mse
    )

    problem = DaisyOptimizationProblem(
        runner, file_generator, objective, parameters, out_dir
    )
    cma_options = {
        "maxfevals" : 500
    }
    with DefaultLogger(out_dir) as logger:
        optimizer = DaisyCMAOptimizer(problem, logger, cma_options)
        result = optimizer.optimize()

    # Optimum at
    # param = 0.1
    # clay = 2 # (But the effect is very small, so dont expect to get this)
    for name, res in result.items():
        print(name)
        for k,v in res.items():
            print('    ', k, ' : ', v, sep='')

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('daisy_path', type=str, help='Path to daisy binary')
    parser.add_argument('--daisy-home', type=str,
                        help='Path to daisy home directory containing lib/ and sample/',
                        default=None)
    args = parser.parse_args()
    single_objective_with_python(args.daisy_path, args.daisy_home)



from ax.api.configs import RangeParameterConfig, ChoiceParameterConfig
from .parameter import ContinuousParameter, CategoricalParameter

def daisy_param_to_ax_param(param):
    '''Convert Daisy parameter to Ax parameter

    Parameters
    ----------
    param : ContinuousParameter OR CategoricalParameter

    Returns
    -------
    RangeParameterConfig OR ChoiceParameterConfig
    '''
    if isinstance(param, ContinuousParameter):
        return RangeParameterConfig(
            name=param.name, parameter_type='float', bounds=param.valid_range
        )
    if isinstance(param, CategoricalParameter):
        init_value = param.values[param.initial_valud_idx]
        if isinstance(init_value, int):
            p_type = 'int'
        elif isinstance(init_value, float):
            p_type = 'float'
        elif isinstance(init_value, str):
            p_type = 'str'
        elif isinstance(init_value, bool):
            p_type = 'bool'
        else:
            raise ValueError('param values must be of type int, float, str or bool')
        return ChoiceParameterConfig(
            name=param.name, parameter_type=p_type, values=param.values, is_ordered=False)
    raise ValueError('param must of type ContinuousParameter or CategoricalParameter')



from abc import ABC, abstractmethod

class Log(ABC):
    '''Log interface'''
    @abstractmethod
    def log(self, *args, **kwargs):
        '''Handle any kind of log mesage'''

    @abstractmethod
    def close(self):
        '''Explicitly close the log'''

    @abstractmethod
    def __del__(self):
        pass

    @abstractmethod
    def __enter__(self):
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



'''Objective functors'''
# pylint: disable=unused-import
from daisypy.optim.scalar_objective import ScalarObjective
from daisypy.optim.aggregate_objective import AggregateObjective
from daisypy.optim.multi_objective import MultiObjective



# pylint: disable=too-few-public-methods
from abc import ABC, abstractmethod

class FileGenerator(ABC):
    """Generate a file in the given output directory using the given params"""
    @abstractmethod
    def __call__(self, output_directory, params):
        """
        Parameters
        ----------
        output_directory : str
          Directory to store the generated file in

        params : dict
          Dictionary of parameters.

        Returns
        -------
        paths : { str : str }
          Map from generator type to generated file paths
        """



# pylint: disable=too-many-locals,R0801
"""Example showing how to combine multiple outputs"""
import argparse
from pathlib import Path
import pandas as pd
from daisypy.optim import (
    ContinuousParameter,
    DefaultLogger,
    DaisyCMAOptimizer,
    DlfDataExtractor,
    DaiFileGenerator,
    MultiFileGenerator,
    DaisyRunner,
    DaisyOptimizationProblem,
    ScalarObjective,
    mse
)

# The function that combines log files is called with a list of DataFrames with a 'time' column
# and some value columns.
# It should return a single DataFrame with columns 'time' and 'value'
# And it has to be defined at the top-level because we cannot pass local functions due to
# multiprocessing.
def combine_logs(data_frames):
    '''
    Combine outputs by summing "Residuals-*" in field_nitrogen.dlf and multiplying by
    "Soil matrix water" in field_water.dlf

    '''
    df = data_frames[0]
    for df2 in data_frames[1:]:
        df = pd.merge(df, df2, on='time', validate='1:1')
    # The name of value columns follow this pattern <log-name>/<variable-name>
    # This value does not really makes sense, but it illustrates how we can combine outputs
    value = df['field_water.dlf/Soil matrix water'] * (
        df['field_nitrogen.dlf/Residuals-Soil'] + df['field_nitrogen.dlf/Residuals-Surface']
    )
    return pd.DataFrame({'time' : df['time'], 'value' : value })

def single_objective_combined_outputs(daisy_path, daisy_home=None):
    '''Optimize by combining multiple outputs

    daisy_path: str
      Path to daisy binary

    daisy_home: str
      Path to daisy home. If None let the Daisy binary figure it out
    '''
    base_dir = Path(__file__).parent
    out_dir = base_dir / 'out' / 'single-objective-combined-outputs'
    data_dir = base_dir / 'example-data' / 'combined-outputs'
    file_generator = MultiFileGenerator({
        'dai' : DaiFileGenerator('run.dai', template_file_path=data_dir / 'template.dai'),
    })

    data_extractor = DlfDataExtractor({
        'field_water.dlf' : 'Soil matrix water',
        'field_nitrogen.dlf' : ['Residuals-Soil', 'Residuals-Surface']
    }, combine_logs)

    parameters = {
        'dai' : [
            ContinuousParameter('fertilize_weight_1', 20, (0, 200)),
            ContinuousParameter('fertilize_weight_2', 20, (0, 200))
        ],
    }

    runner = DaisyRunner(daisy_path, daisy_home)

    target_file = data_dir / 'target.csv'
    objective = ScalarObjective(
        name="objective",
        data_extractor=data_extractor,
        target=target_file,
        target_name="target",
        loss_fn=mse
    )

    problem = DaisyOptimizationProblem(
        runner, file_generator, objective, parameters, out_dir
    )
    cma_options = {
        "maxfevals" : 200
    }
    with DefaultLogger(out_dir) as logger:
        optimizer = DaisyCMAOptimizer(problem, logger, cma_options)
        result = optimizer.optimize()

    # Optimum at
    for name, res in result.items():
        print(name)
        for k,v in res.items():
            print('    ', k, ' : ', v, sep='')
    print('Optimum at',
          '  fertilize_weight_1 = 100',
          '  fertilize_weight_2 = 80',
          sep='\n')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('daisy_path', type=str, help='Path to daisy binary')
    parser.add_argument('--daisy-home', type=str,
                        help='Path to daisy home directory containing lib/ and sample/',
                        default=None)
    args = parser.parse_args()
    single_objective_combined_outputs(args.daisy_path, args.daisy_home)



# pylint: disable=too-many-locals,too-few-public-methods,R0801
"""Multi objective optimization over multiple scenarios"""
import argparse
from pathlib import Path
import pandas as pd
from daisypy.optim import (
    DaiFileGenerator,
    DlfDataExtractor,
    ScalarObjective,
    MultiObjective,
    DaisyOptimizationProblem,
    DaisyAxOptimizer,
    ContinuousParameter,
    DaisyRunner,
    DefaultLogger
)

# We use the multiprocessing module, which uses pickle, so we cannot use local functions for loss
# functions and aggregate functions
def mse(actual, target):
    """Mean squared error"""
    return ((actual - target)**2).mean()

def multi_objective_ax(daisy_path):
    '''Multi objective optimization using DaisyAxOptimizer'''
    base_dir = Path(__file__).parent
    out_dir = base_dir / 'out' / 'multi-objective-ax'
    data_dir = base_dir / 'example-data' / 'multiple-scenarios'

    # 0. Define a runner that can run Daisy
    runner = DaisyRunner(daisy_path)

    # 1. Setup the dai file generator
    dai_template = data_dir / 'template.dai'
    dai_file_generator = DaiFileGenerator(template_file_path=dai_template)

    # 2. Define the parameters that we will optimize
    # Names of parameters should match the names in the template file
    parameters = [
        ContinuousParameter(
            name='temp_offset',
            initial_value=0,
            valid_range=(-2, 2)
        ),
    ]

    # 3. Define the objective
    # We have three scenarios. These where generated with
    # Askov    : temp_offset =  0 (366 samples
    # Jyndevad : temp_offset = -2 (307 samples)
    # Foulum   : temp_offset =  2 (215 samples)
    scenarios = [ 'askov', 'jyndevad', 'foulum' ]
    targets = { name : pd.read_csv(data_dir / f'target-{name}.csv') for name in scenarios }
    target_names = { k : 'Leaching' for k in scenarios }

    # The logs do not have to be the same
    log_names = { k : f'{k}/field_nitrogen.dlf' for k in scenarios }

    # The variables also dont have to be the same
    variables = { k : 'Matrix-Leaching' for k in scenarios }

    # And the losses do not have to be the same
    losses = { k : mse for k in scenarios }
    objective_fns = [
        ScalarObjective(
            name,
            DlfDataExtractor({log_names[name] : variables[name]}),
            targets[name],
            target_names[name],
            losses[name]
        ) for name in scenarios
    ]
    objective = MultiObjective('multi', objective_fns)


    # 4. Wrap everything as an optimization problem
    # Normally we would not set data_dir and we would set debug = False,
    # but here we set them so we can inspect the output.
    # If debug = False, then outputs are deleted as soon as the optimizer is done with them
    out_data_dir = out_dir / 'data_dir'
    debug = True
    problem = DaisyOptimizationProblem(
        runner, dai_file_generator, objective, parameters, out_data_dir, debug
    )

    # 5. Setup a logger
    # We use DefaultLogger that logs parameter distributions and sampled parameters to csv files
    log_dir = out_dir / 'logs'
    logger = DefaultLogger(log_dir)

    # 6. Setup an optimizer
    options = {
        "max_trials" : 25,
        "max_trials_iteration" : 3
    }
    optimizer = DaisyAxOptimizer(problem, logger, options)
    assert optimizer.multi_objective, "Optimizer is not multi objective"

    # 7. Run the optimizer
    results = optimizer.optimize()

    # 8. Look at the results
    for result in results:
        for param, value in result.parameters.items():
            print(param, value)
        for metric, value in result.metrics.items():
            print(metric, value)
        print('--------------------------------------------------------------------------------')



if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('daisy_path', type=str, help='Path to daisy binary')
    args = parser.parse_args()
    multi_objective_ax(args.daisy_path)



import os
from .log import Log
from .formatters import quote_if_string

class CsvLog(Log):
    '''A file backed csv log.'''

    def __init__(self, path, columns=None, default_formatter=quote_if_string):
        '''
        Parameters
        ----------
        path: str
          Where to store log

        columns : list of string or dict of (str, callable)
          Column names and formats. If a list use default format function for all columns (`str`).
          If None determine columns from first call to self.log

        default_formatter : callable [Object -> str]
          Default function for formatting column values. If None use self.quote_if_string
        '''
        _dir = os.path.dirname(path)
        os.makedirs(_dir, exist_ok=True)
        self.path = path
        self._log = open(path, 'w', encoding='utf-8') # pylint: disable=consider-using-with
        if columns is None:
            # Deferred setting of columns such that they can be set on first write
            self.columns = None
            self.default_formatter = default_formatter
        else:
            self._setup_columns(columns, default_formatter)

    def log(self, *args, flush=True, **kwargs):
        '''Log a row.

        Parameters
        ----------
        flush : Bool
          If True flush the log after writing.

        **kwargs : dict
          If the log has a column specification, then this dict must contain the columns defined
          there. Otherwise the dict keys are used to create a column specification.
        '''
        if self._log.closed:
            raise RuntimeError('Writing to closed CsvLog')
        if len(args) > 0:
            raise RuntimeError('CsvLog requires all log arguments to be passed as keywords')
        if self.columns is None:
            self._setup_columns(list(kwargs.keys()))
        row = []
        for col, formatter in self.columns.items():
            row.append(formatter(kwargs[col]))
        self._write(','.join(row), flush)

    def close(self):
        '''Close the underlying file'''
        self._log.close()

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _write(self, msg, flush):
        self._log.write(msg + "\n")
        if flush:
            self.flush()

    def flush(self):
        '''Flush the log so it is written to disk'''
        self._log.flush()
        os.fsync(self._log.fileno())

    def _setup_columns(self, columns, default_formatter=None):
        if default_formatter is None:
            default_formatter = self.default_formatter
        if not isinstance(columns, dict):
            self.columns = { col : default_formatter for col in columns }
        else:
            self.columns = {
                k : default_formatter if v is None else v for k,v in columns.items()
            }
        self._write(','.join(self.columns.keys()), True)



import argparse
import os
import time
from pathlib import Path
import pandas as pd
from generate_time_series import generate_time_series # pylint: disable=import-error

def char_range(n):
    '''Generate a range of characters'''
    return [chr(ord('a') + i) for i in range(n)]

def generate_benchmark_data():
    '''Generate varying sized csv files for benchmarking purposes'''
    data_dir = Path(__file__).parent / 'benchmark-data'
    os.makedirs(data_dir, exist_ok=True)
    print('Generating small', flush=True)
    small = generate_time_series('2000-01-01', '2020-01-01', 200, char_range(2))
    small.to_csv(data_dir / 'small.csv')

    print('Generating medium', flush=True)
    medium = generate_time_series('2000-01-01', '2020-01-01', 2000, char_range(5))
    medium.to_csv(data_dir / 'medium.csv')

    print('Generating large', flush=True)
    large = generate_time_series('2000-01-01', '2020-01-01', 20000, char_range(20))
    large.to_csv(data_dir / 'large.csv')

    print('Generating very large', flush=True)
    very_large = generate_time_series('2000-01-01', '2020-01-01', 200000, char_range(50))
    very_large.to_csv(data_dir / 'very-large.csv')

def benchmark_read_csv():
    '''Compare pandas.read_csv using C engine with sep=',' and python engine with sep=None'''
    in_dir = Path(__file__).parent / 'benchmark-data'
    names = ['small', 'medium', 'large', 'very-large']
    data_files = [ in_dir / f'{name}.csv' for name in names ]

    c_timings = []
    for data_file in data_files:
        print(f'Start C {data_file}', flush=True)
        start = time.perf_counter()
        df = pd.read_csv(data_file, sep=',', engine='c') # pylint: disable=unused-variable
        elapsed = time.perf_counter() - start
        c_timings.append(elapsed)

    py_timings = []
    for data_file in data_files:
        print(f'Start python {data_file}', flush=True)
        start = time.perf_counter()
        df = pd.read_csv(data_file, sep=None, engine='python') # pylint: disable=unused-variable
        elapsed = time.perf_counter() - start
        py_timings.append(elapsed)


    results = pd.DataFrame({'name' : names, 'C' : c_timings, 'python' : py_timings})
    results['python/C'] = results['python'] / results['C']
    print(results)

def check_data_exists():
    '''Check that all benchmarking data exists'''
    in_dir = Path(__file__).parent / 'benchmark-data'
    names = ['small', 'medium', 'large', 'very-large']
    data_files = [ in_dir / f'{name}.csv' for name in names ]
    for data_file in data_files:
        if not data_file.exists():
            return False
    return True

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--regenerate', action='store_true', help='If set regenerate benchmark data'
    )
    args = parser.parse_args()
    if args.regenerate or not check_data_exists():
        print('Generating data')
        generate_benchmark_data()
    print('Running benchmarks')
    benchmark_read_csv()



# pylint: disable=too-few-public-methods,R0801
from concurrent.futures import ProcessPoolExecutor
import numpy as np
from .parameter import CategoricalParameter
from .problem import ScalarProblemWrapper

class DaisySequentialOptimizer:
    """Daisy optimizer using a sequential approach

    The method starts from the initial parameters. Then it changes each parameter in turn.
    The single parameter leading to best performance is then fixed and the process repeated
    untill all parameters are fixed.
    """
    def __init__(self, problem, logger, options=None, number_of_processes=None):
        """
        Parameters
        ----------
        problem : DaisyProblem

        logger : ...

        options : dict

        number_of_processes: int > 0 (Optional)
          The maximum number of processes to use when running Daisy. Defaults to
          os.process_cpu_count()
        """
        if options is None:
            options = {}
        self.objective_name = problem.objective_fn.name
        self.problem = ScalarProblemWrapper(problem)
        self.logger = logger
        self.number_of_processes = number_of_processes

        # Convert any continuous parameters to categorical parameters by uniform sampling
        num_samples = options.get("num_samples", 3)
        self.parameters = []
        for param in problem.parameters:
            # Standardize parameters such that they are all categorical and the initial value is the
            # first value in the values list.
            if param.type == "Continuous":
                # If num_samples == 2, then only the initial and lower end of the valid range is
                # used.
                # We can end up with the initial_value being added twice, if for example we have
                # valid_range = (0, 10)
                # initial_value = 5
                # num_samples = 12
                # It is not clear what is the best solution to this. If we sample the ranges
                # (low, initial) and (initial, high) separately then we get different spacing in
                # most cases.
                # For now we just ignore it and accept that we sometimes try one less parameter than
                # we would like.
                values = np.concatenate([
                    [param.initial_value],
                    np.linspace(param.valid_range[0], param.valid_range[1], num_samples-1)
                ])
                self.parameters.append(CategoricalParameter(param.name, values, 0))
            elif param.type == 'Categorical':
                if param.initial_value_idx != 0:
                    values = np.concatenate([
                        [param.values[param.initial_value_idx]],
                        param.values[:param.initial_value_idx],
                        param.values[param.initial_value_idx+1:]
                    ])
                    param = CategoricalParameter(param.name, values, 0)
                self.parameters.append(param)

    def optimize(self):
        '''Run optimization'''
        # pylint: disable=too-many-locals,too-many-statements
        # Recall that we are working with categorical parameters, so there is no sampling of new
        # parameters.
        step = 0
        fixed = set()   # The parameters that are already fixed
        floating = {}   # The parameters that we need to fix
        current = {}    # The parameter values that we are currently using
        order = []      # Order that parameters are passed to the problem.
        num_param_values = [] # Number of possible parameter values for each parameter
        for param in self.parameters:
            floating[param.name] = param.values
            current[param.name] = param.values[0] # Parameter values are tried in order
            order.append(param.name)
            num_param_values.append(len(param.values))

        min_evals, max_evals = _count_min_max_param_evals(num_param_values)
        self.logger.info(f'Using at least {min_evals} and at most {max_evals} function evaluations')

        # Compute the initial loss
        self.logger.info('Evaluating initial parameters')
        current_fval = self.problem([current[name] for name in order])
        if np.isnan(current_fval):
            self.logger.error('Initial parameters failed, aborting')
            raise RuntimeError('Initial parameters failed')

        self.logger.info(f'Initial objective = {current_fval}')
        total_f_evals = 1
        self.logger.info('Optimizing')
        with ProcessPoolExecutor(self.number_of_processes) as executor:
            while len(floating) > 0:
                # We fix a parameter in each step, so we will always do as many steps as there are
                # parameters.
                step += 1

                # Log the parameter distribution
                params = {
                    f'param_{name}_choices' : ','.join([str(v) for v in values])
                    for name, values in floating.items()
                }
                for name in fixed:
                    params[f'param_{name}_choices'] = str(current[name])
                self.logger.parameters(distribution='categorical', tag='raw', step=step, **params)

                param_sets, param_sets_ids = _generate_parameter_sets(floating, current, order)
                self.logger.info(step=step, n_param_sets=len(param_sets))

                best = np.inf
                best_idx = None
                num_failures = 0
                # executor.map runs the problems in parallel and yields results in order matching
                # param_sets.
                for i, fval in enumerate(executor.map(self.problem, param_sets)):
                    objective_value = { f'metric_{self.objective_name}' : fval }
                    params = {
                        f'param_{name}' : value for name, value in zip(order, param_sets[i])
                    }
                    self.logger.result(
                        step=step, tag="raw", **objective_value, **params
                    )
                    if np.isnan(fval):
                        num_failures += 1
                    elif fval < best:
                        best = fval
                        best_idx = i # Index into param_sets
                if best_idx is None:
                    # Maybe not raise an exception if we have had at least one successful run in a
                    # previous step?
                    self.logger.error('All simulations failed. Aborting')
                    raise RuntimeError('All simulations failed')

                total_f_evals += len(param_sets)
                self.logger.info(step=step, total_function_evaluations=total_f_evals)
                if num_failures > 0:
                    self.logger.warning(step=step, n_failed_runs=num_failures)
                self.logger.info(step=step, best_objective=best)
                if best > current_fval:
                    # Nothing is better than using current values of all parameters, so we stop.
                    # We could consider setting a random parameter to a random value, or something
                    # similar.
                    self.logger.info('No improvement in objective. Stopping')
                    break

                current_fval = best
                name, idx = param_sets_ids[best_idx]
                value = floating.pop(name)[idx]
                current[name] = value
                fixed.add(name)
                self.logger.info(f'step={step},Fixing {name} to {value}')

        result = {}
        for k,v in current.items():
            result[k] = { 'best': v }
        return result

def _count_min_max_param_evals(num_param_values):
    # Count the minimum and maximum number of function evaluations
    # Worst case is that we always fix the parameter with fewest values
    num_param_values = sorted(num_param_values)
    max_evals = 1 # We always do one with the current parameter set
    for start in range(len(num_param_values)):
        for n in num_param_values[start:]:
            max_evals += n-1 #

    # Best case is that we always fix the parameter with most values
    min_evals = 1
    num_param_values = num_param_values[::-1]
    for start in range(len(num_param_values)):
        for n in num_param_values[start:]:
            min_evals += n-1
    return min_evals, max_evals

def _generate_parameter_sets(floating, current, order):
    # Generate parameter sets where all parameters, exept one, are fixed
    # A parameter set is a dict of (parameter name, parameter value)
    # For a specific parameter p, we keep all other parameters fixed and then generate
    # parameter sets where p is varied over all its possible values
    param_sets_ids = []
    param_sets = []
    for name, values in floating.items():
        # We keep all except name fixed
        for i, value in enumerate(values):
            if value == current[name]:
                # This is the case where all parameters have their current value and we
                # have already computed this combination in the previous step regardless
                # of which parameter was fixed
                continue

            param_set = []
            for param_name in order: # We must maintain the order of parameters
                if param_name == name:
                    param_set.append(float(value))
                else:
                    param_set.append(float(current[param_name]))
            param_sets.append(param_set)
            param_sets_ids.append((name, i))
    return param_sets, param_sets_ids



# pylint: disable=missing-function-docstring
from pathlib import Path
from daisypy.optim.py_file_generator import PyFileGenerator

EXPECTED = """def linear(x):
    return 0.5 * x + 10
"""
PARAMS = {'a' : 0.5, 'b' : 10}

def test_tagged(tmp_path):
    template_path = Path(__file__).parent / 'templates' / 'template.py'
    generator = PyFileGenerator('testing.py', template_file_path=template_path)
    file_path = Path(generator(tmp_path, {'py' : PARAMS})['py'])
    assert file_path.read_text(encoding='utf-8') == EXPECTED

def test_not_tagged(tmp_path):
    template_path = Path(__file__).parent / 'templates' / 'template.py'
    generator = PyFileGenerator('testing.py', template_file_path=template_path)
    file_path = Path(generator(tmp_path, PARAMS, tagged=False))
    assert file_path.read_text(encoding='utf-8') == EXPECTED

def test_no_params(tmp_path):
    template = "x = {{ 'a' : 1 }}"
    expected = "x = { 'a' : 1 }"
    generator = PyFileGenerator('testing.py', template_text=template)
    file_path = Path(generator(tmp_path, {}, tagged=False))
    assert file_path.read_text(encoding='utf-8') == expected
    file_path = Path(generator(tmp_path, {'py' : {}})['py'])
    assert file_path.read_text(encoding='utf-8') == expected



import pandas as pd
from .loss_wrapper import LossWrapper

class ScalarObjective:
    # pylint: disable=too-few-public-methods,too-many-arguments,too-many-positional-arguments
    """Scalar objective that extracts data from a daisy output directory and computes a loss"""

    def __init__(self, name, data_extractor, target, target_name, loss_fn):
        """
        Parameters
        ----------
        name : str
          Name of objective

        data_extractor : DlfDataExtractor
          Extractor mapping output directories to pandas.Series

        target : pandas.DataFrame OR str
          If str it is opened with pandas.read_csv.
          Must contain columns "time" and `target_name`

        target_name : str
          Name of column in target that contains the target values

        loss_fn : callable : (actual, target) -> loss
          The loss function to use
        """
        self.name = name
        self.data_extractor = data_extractor
        if not isinstance(target, pd.DataFrame):
            target = pd.read_csv(target, sep=None, engine='python')
        if not "time" in target.columns:
            raise ValueError(
                f'target must contain "time" column. Got columns {list(target.columns)}'
            )
        if not target_name in target.columns:
            raise ValueError(
                f'target must contain "{target_name}" column. Got columns {list(target.columns)}'
            )
        self.target = target[["time", target_name]].rename(columns={target_name : 'value'})
        self.target["time"] = pd.to_datetime(self.target["time"])
        self.loss_fn = LossWrapper(loss_fn) # Wrap it so target and actual are processed correctly

    def __call__(self, daisy_output_directory):
        """Compute the objective

        Parameters
        ----------
        daisy_output_directory : str
          Path to daisy ouput directory

        Returns
        -------
        objective_map : dict of [str, float]
          Map from the objective name to the objective value
        """
        actual = self.data_extractor(daisy_output_directory)
        return { self.name : self.loss_fn(actual, self.target) }



# pylint: disable=too-many-locals,too-few-public-methods,R0801
"""Example showing how to optimize two Daisy parameters over multiple scenarios"""
import argparse
from pathlib import Path
import numpy as np
import pandas as pd
from daisypy.optim import (
    AggregateObjective,
    DaiFileGenerator,
    DlfDataExtractor,
    DaisySequentialOptimizer,
    ScalarObjective,
    DaisyOptimizationProblem,
    ContinuousParameter,
    DaisyRunner,
    DefaultLogger
)

# We use the multiprocessing module, which uses pickle, so we cannot use local functions for loss
# functions and aggregate functions
def mse(actual, target):
    """Mean squared error"""
    return ((actual - target)**2).mean()

class WeightedAverage:
    '''Compute weighted average'''
    def __init__(self, weights):
        self.w = np.array(weights)

    def __call__(self, x):
        return (list(x.values()) * self.w).sum()

def multiple_scenarios(daisy_path):
    '''How to optimize parameters for Daisy

    0. Define a runner that can run Daisy
    1. Setup the dai file generator
    2. Define the parameters that we will optimize
    3. Define the objective
    4. Wrap everything as an optimization problem
    5. Setup a logger
    6. Choose an optimizer
    7. Run the optimizer
    8. Look at the results
    '''
    base_dir = Path(__file__).parent
    out_dir = base_dir / 'out' / 'multiple-scenarios'
    data_dir = base_dir / 'example-data' / 'multiple-scenarios'

    # 0. Define a runner that can run Daisy
    runner = DaisyRunner(daisy_path)

    # 1. Setup the dai file generator
    dai_template = data_dir / 'template.dai'
    dai_file_generator = DaiFileGenerator(template_file_path=dai_template)

    # 2. Define the parameters that we will optimize
    # Names of parameters should match the names in the template file
    parameters = [
        ContinuousParameter(
            name='temp_offset',
            initial_value=0,
            valid_range=(-2, 2)
        ),
    ]

    # 3. Define the objective
    # We have three scenarios. We want to optimize the mean squared error weighted by the number
    # of measurements in each target
    # These where generated with
    # Askov    : temp_offset =  0 (366 samples
    # Jyndevad : temp_offset = -2 (307 samples)
    # Foulum   : temp_offset =  2 (215 samples)
    scenarios = [ 'askov', 'jyndevad', 'foulum' ]
    targets = { name : pd.read_csv(data_dir / f'target-{name}.csv') for name in scenarios }
    target_names = { k : 'Leaching' for k in scenarios }

    # The logs do not have to be the same
    log_names = { k : f'{k}/field_nitrogen.dlf' for k in scenarios }

    # The variables also dont have to be the same
    variables = { k : 'Matrix-Leaching' for k in scenarios }

    # And the losses do not have to be the same
    losses = { k : mse for k in scenarios }
    objective_fns = [
        ScalarObjective(
            name,
            DlfDataExtractor({log_names[name] : variables[name]}),
            targets[name],
            target_names[name],
            losses[name]
        ) for name in scenarios
    ]

    # Define the weighting
    # We could also do this using sum of squared error as loss_fn and aggregate with mean, but this
    # way illustrates how we can easily define different weights.
    # Just be careful that the order of weights match the order of the targets passed to
    # AggregateObjective
    # This should give temp_offset = -1.5
    weights = np.array(list(map(len, targets)), dtype='float32')
    weights /= weights.sum()

    # You can also try these.
    # only_askov = [1, 0, 0]
    # only_jyndevad = [0, 1, 0]
    # only_foulum = [0, 0, 1]
    # We should get
    #  Askov : temp_offset = 0,
    #  jyndevad : temp_offset = -2
    #  foulum : temp_offset = 2


    # You could also compute weights based on variance of targets
    # This should give temp_offset = -2
    # var_weights = np.array([np.var(target[variable]) for target in targets])
    # var_weights /= var_weights.sum()

    aggregate_fn = WeightedAverage(weights)
    objective = AggregateObjective('aggregated', objective_fns, aggregate_fn)


    # 4. Wrap everything as an optimization problem
    # Normally we would not set data_dir and we would set debug = False,
    # but here we set them so we can inspect the output.
    # If debug = False, then outputs are deleted as soon as the optimizer is done with them
    out_data_dir = out_dir / 'data_dir'
    debug = True
    problem = DaisyOptimizationProblem(
        runner, dai_file_generator, objective, parameters, out_data_dir, debug
    )

    # 5. Setup a logger
    # We use DefaultLogger that logs parameter distributions and sampled parameters to csv files
    log_dir = out_dir / 'logs'
    logger = DefaultLogger(log_dir)

    # 6. Setup an optimizer
    # We use sequential because CMA does not work well with a single parameter (it's an
    # implementation thing)
    options = {
        'num_samples' : 18
    }
    optimizer = DaisySequentialOptimizer(problem, logger, options)

    # 7. Run the optimizer
    result = optimizer.optimize()

    # 8. Look at the results
    for name, res in result.items():
        print(name)
        for k,v in res.items():
            print('    ', k, ' : ', v, sep='')

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('daisy_path', type=str, help='Path to daisy binary')
    args = parser.parse_args()
    multiple_scenarios(args.daisy_path)



# pylint: disable=wrong-import-position
available_optimizers = {}
try:
    from daisypy.optim.cma_optimizer import DaisyCMAOptimizer
    available_optimizers["cma"] = DaisyCMAOptimizer
except ImportError:
    pass

try:
    from daisypy.optim.ax_optimizer import DaisyAxOptimizer, AxResult #pylint: disable=unused-import
    available_optimizers["ax"] = DaisyAxOptimizer
except ImportError:
    pass

from daisypy.optim.sequential_optimizer import DaisySequentialOptimizer
available_optimizers["sequential"] = DaisySequentialOptimizer



# pylint: disable=too-few-public-methods
from subprocess import CompletedProcess
from daisypy.optim.file_generator import FileGenerator

class MockFileGenerator(FileGenerator):
    '''Mock file generator that always generates the paths it was constructed with'''
    def __init__(self, paths):
        self.paths = paths

    def __call__(self, output_directory, params, tagged=True):
        return self.paths


class MockRunner:
    '''Mock runner always returning a CompletedProcess with a specified returncode'''
    def __init__(self, args=None, returncode=0):
        self.args = args if args is not None else []
        self.returncode = returncode

    def __call__(self, dai_file, output_directory):
        return CompletedProcess(self.args, self.returncode)

class MockObjective:
    '''Mock objective always returning a specific value'''
    def __init__(self, name='mock', value=0):
        self.name = name
        self.value = value

    def __call__(self, daisy_output_directory):
        return { self.name : self.value }


class MockProblem:
    '''Problem case for test purposes. Will evaluate an objective by forwarding parameters'''
    def __init__(self, parameters, objective_fn):
        self.parameters = parameters
        self.objective_fn = objective_fn

    def __call__(self, parameter_values):
        named_parameters = { p.name : value for p, value in zip(self.parameters, parameter_values) }
        return { 'mock' : self.objective_fn(**named_parameters) }

class MockDataExtractor:
    '''Mock data extractor returning data it was constructed with'''
    def __init__(self, data):
        self.data = data

    def __call__(self, daisy_output_directory):
        return self.data

class MockLoss:
    '''Mock loss that always returns a specificed value'''
    def __init__(self, value):
        self.value = value

    def __call__(self, actial, target):
        return self.value



class Logger:
    """A logger that handles a number of logs.
    Use as context manager OR call close method explicitly OR ensure the object is destroyed
    """
    def __init__(self, **kwargs):
        self.logs = kwargs

    def add_log(self, name, log):
        '''Add a named log

        Parameters
        ----------
        name : str
          Name of log

        log : daisypy.optim.Log
          A Log instance

        Raises
        ------
        ValueError if a log with the given name already exists
        '''
        if name in self.logs:
            raise ValueError(f"Log {name} already exists")
        self.logs[name] = log

    def info(self, *args, **kwargs):
        '''Log to the info log'''
        self.log('info', *args, **kwargs)

    def warning(self, *args, **kwargs):
        '''Log to the warning log'''
        self.log('warning', *args, **kwargs)

    def error(self, *args, **kwargs):
        '''Log to the error log'''
        self.log('error', *args, **kwargs)

    def parameters(self, *args, **kwargs):
        '''Log to the parameters log'''
        self.log('parameters', *args, **kwargs)

    def result(self, *args, **kwargs):
        '''Log to the result log'''
        self.log('result', *args, **kwargs)

    def log(self, name, *args, **kwargs):
        '''Log to a specific named log. Fallback to `default` log if the named log does not exist

        Parameters
        ----------
        name : str
          Name of log

        Raises
        ------
        ValueError if no log with the given name exists AND no default log is defined
        '''
        if name not in self.logs:
            if 'default' in self.logs:
                self.logs['default'].log(*args, **kwargs)
            else:
                raise ValueError(f"No log named {name} exists")
        else:
            self.logs[name].log(*args, **kwargs)

    def close(self):
        '''Close all logs for writing'''
        for log in self.logs.values():
            log.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()



# pylint: disable=too-few-public-methods
'''Test functions from https://en.wikipedia.org/wiki/Test_functions_for_optimization'''
from daisypy.optim import ContinuousParameter

class BealeFunction:
    '''The BealeFunction'''
    def __init__(self):
        self.name = "Beale"
        self.parameters = [
            ContinuousParameter('x', 0, (-4,4)),
            ContinuousParameter('y', 0, (-4,4))
        ]
        self.amin = { 'x' : 3, 'y' : 0.5 }
        self.min = 0

    def __call__(self, x, y):
        return (1.5 - x + x*y)**2 + (2.25 - x + x*y**2)**2 + (2.625 - x + x*y**3)**2

beale_function = BealeFunction()



'''Module for Daisy parameter optimization'''
from daisypy.optim._version import version
from daisypy.optim.aggregate_fns import *
from daisypy.optim.file_generators import *
from daisypy.optim.logging import *
from daisypy.optim.loss_fns import *
from daisypy.optim.objective import *
from daisypy.optim.optimizer import *
from daisypy.optim.parameter import *
from daisypy.optim.problem import DaisyOptimizationProblem
from daisypy.optim.runner import DaisyRunner
from daisypy.optim.visualize import *
from daisypy.optim.dlf_data_extraction import (
    DlfDataExtractor,
    DlfPostProcessor,
    DlfSum
)



# pylint: disable=too-few-public-methods
'''Classes for extracting and procesing data in Daisy log files'''
from abc import ABC, abstractmethod
from collections.abc import Mapping
from pathlib import Path
import pandas as pd
from daisypy.io.dlf import read_dlf

class DlfDataExtractor:
    '''Class for extracting data from Daisy log files (dlf)'''
    def __init__(self, logs_and_variables, post_processor=None):
        '''
        Parameters
        ----------
        logs_and_variables : dict of [str, [str]]
          Map from log names to variable names in the log that should be extracted

        post_processor : Callable implementing DlfPostProcessor interface or None
          Function mapping extracted data to a single pandas.DataFrame with columns 'time' and
          'value'. It is called with a list of pd.DataFrames with a "time" column and one or more
          value columns named with '<log-name>/<var-name>'.
          If None then logs_and_variables must contain exactly one key, and that key must map to a
          list of length 1 or a single str
        '''
        if not isinstance(logs_and_variables, Mapping):
            raise ValueError("`logs_and_variables` must be a Mapping type")
        self.logs_and_variables = logs_and_variables
        if post_processor is None:
            keys = list(self.logs_and_variables.keys())
            msg = ("When no `post_processor` is provided, `logs_and_variables` must contain "
                "exactly one key, and that key must map to a list of length 1 or a single str")
            if len(keys) != 1:
                raise ValueError(msg)
            val = self.logs_and_variables[keys[0]]
            if not isinstance(val, str) and len(val) != 1:
                raise ValueError(msg)
            self.post_processor = DlfSingleton()
        else:
            self.post_processor = post_processor

    def __call__(self, daisy_output_directory):
        '''Extract data from log files in the given directory

        Returns
        -------
        pandas.DataFrame with columns 'time' and 'value'
        '''
        out_dir = Path(daisy_output_directory)
        dfs = []
        for log_name, var_names in self.logs_and_variables.items():
            if isinstance(var_names, str):
                var_names = [var_names]
            dlf = read_dlf(out_dir / log_name)
            dlf.body['time'] = pd.to_datetime(
                dlf.body[['year', 'month', 'mday', 'hour']].rename(columns={'mday' : 'day'})
            )
            dfs.append(dlf.body[['time'] + var_names].rename(columns={
                var_name : f'{log_name}/{var_name}' for var_name in var_names
            }))
        processed = self.post_processor(dfs)
        if set(processed.columns) != {'time', 'value'}:
            raise RuntimeError(
                "`post_processor` must return a DataFrame with columns 'time' and 'value'"
            )
        return processed


class DlfPostProcessor(ABC):
    '''Interface for post processors'''
    @abstractmethod
    def __call__(self, data_frames):
        '''
        Parameters
        ----------
        data_frames : [pandas.DataFrame]
          A list of data frames containing a 'time' column and one or more value columns

        Returns
        -------
        pandas.DataFrame with columns 'time' and 'value'
        '''

class DlfSum(DlfPostProcessor):
    '''Compute sum of all extracted variables'''
    def __call__(self, data_frames):
        df = data_frames[0]
        for other in data_frames[1:]:
            df = pd.merge(df, other, on='time', validate='1:1')
        cols = [col for col in df.columns if col != 'time']
        return pd.DataFrame({'time' : df['time'], 'value' : df[cols].sum(axis=1)})

class DlfSingleton(DlfPostProcessor):
    '''Unpack a single extracted variable'''
    def __call__(self, data_frames):
        df = data_frames[0]
        value_col = [col for col in df.columns if col != 'time'][0]
        return df[['time', value_col]].rename(columns={value_col : 'value'})



from daisypy.optim.csv_log import CsvLog

def test_csv_log(tmp_path):
    '''Test that CsvLog logs the expected messages and handles cleanup nicely'''
    expected = [
        'tag,step,value,message',
        'test,1,0.1,"quoted string"',
        'test,2,0.2,"quoted string 2"',
        'done,3,0.3,"quoted string 3"',
    ]
    columns = {
        'tag' : str,
        'step' : lambda x : str(int(x)),
        'value' : lambda x : str(float(x)),
        'message' : None
    }
    path = tmp_path / 'test-log.csv'
    with CsvLog(path, columns) as log:
        log.log(tag='test', step=1, value=0.1, message="quoted string", flush=False)
        log.log(tag='test', step=2, value=0.2, message="quoted string 2", flush=True)
        with open(path, 'r', encoding='utf-8') as infile:
            lines = [line.strip() for line in infile]
        assert lines == expected[:-1]
        log.log(tag='done', step=3, value=0.3, message="quoted string 3", flush=False)

    # Should be flushed when closed
    with open(path, 'r', encoding='utf-8') as infile:
        lines = [line.strip() for line in infile]
    assert lines ==  expected



# pylint: disable=too-few-public-methods
from .file_generator import FileGenerator

class MultiFileGenerator(FileGenerator):
    """Wrapper that handles multiple file generators"""
    def __init__(self, generators):
        """
        Parameters
        ----------
        generators : dict of (str, FileGenerator)
        """
        self.generators = generators

    def __call__(self, output_directory, params, tagged=True):
        """Generate files

        Parameters
        ----------
        output_directory : str
          Directory to store the generated files in

        params : dict of (str, dict)
          Dictionary of parameters. Keys MUST match generator names.
          Keys in nested dictionaries MUST match the corresponding generators parameters

        tagged : bool
          If True return a mapping from generator names to file names otherwise return a list of
          file names

        Returns
        -------
        paths : dict of (str, str) OR list of str
          Map from generator name to generated file OR list of generated files
        """
        paths = {}
        for gen_name, gen_params in params.items():
            paths[gen_name] = self.generators[gen_name](output_directory, gen_params, False)
        if not tagged:
            return list(paths.values())
        return paths



import tempfile
import os
import platform
import numpy as np

class ScalarProblemWrapper:
    # pylint: disable=too-few-public-methods
    '''Helper class that evalues a DaisyOptimizationProblem and extract the value from the returned
    dict'''

    def __init__(self, problem):
        '''
        Parameters
        ----------
        problem : DaisyOptimizationProblem
        '''
        self.problem = problem

    def __call__(self, parameter_values):
        '''
        Parameters
        ----------
        parameter_values : sequence
          Parameter values. Lenght MUST match length of `self.parameters`

        Returns
        -------
        float
        '''
        result = self.problem(parameter_values)
        err_msg = 'Expected a dict with exactly one scalar valued objective mapping'
        try:
            value = result.popitem()[1]
            if len(result) != 0 or not isinstance(value, (int, float)):
                raise RuntimeError(err_msg)
            return value
        except (KeyError, AttributeError, TypeError) as e:
            raise RuntimeError(err_msg) from e


class DaisyOptimizationProblem:
    # pylint: disable=too-many-arguments,too-many-positional-arguments,too-few-public-methods
    '''Class that knows how to run simulation and compute objective for a parameter set'''
    def __init__(
            self, runner, file_generator, objective_fn, parameters, data_dir=None, debug=False
    ):
        """
        Parameters
        ----------
        runner : DaisyRunner

        file_generator : FileGenerator

        objective_fn : DaisyObjective

        parameters : [DaisyParameter] OR dict of (str, [DaisyParameter])
          Parameters to optimize. If a list it is assumed that all parameters are for the 'dai' file

        data_dir : str
          If not None then temporary directories will be created in this directory. Otherwise, they
          will be created in a default location depending on platform.

        debug: bool
          If True do not delete the temporary directory where Daisy output is stored
        """
        self.runner = runner
        self.file_generator = file_generator
        self.objective_fn = objective_fn
        self.parameter_kind = {}
        if not isinstance(parameters, dict):
            parameters = { 'dai' : parameters }

        # Convert dict of parameters to a list of parameters and verify that there are no name
        # clashes.
        self.parameters = []
        for kind, params in parameters.items():
            for param in params:
                if param.name in self.parameter_kind:
                    raise ValueError(
                        'Parameters must have unique names across all templates. '
                        f'{param.name} from {kind} is also in {self.parameter_kind[param.name]}'
                    )
                self.parameter_kind[param.name] = kind
                self.parameters.append(param)

        self.data_dir = data_dir
        if data_dir is None and platform.system().lower() == 'linux':
            # There is a good chance that we are using flatpak, in which case we need to use a tmp
            # location that flatpak Daisy can read and write. Anything inside the user home is good.
            self.data_dir = os.path.expanduser('~/.tmp/daisy')
        if self.data_dir is not None:
            os.makedirs(self.data_dir, exist_ok=True)
        self.debug = debug

    def __call__(self, parameter_values):
        # TODO: Rewrite to accept a dict of parameters. This is too brittle
        """Run Daisy with the given parameters and evaluate the objective. The return value depends
        on

        Parameters
        ----------
        parameter_values : sequence
          Parameter values. Lenght MUST match length of `self.parameters`

        Returns
        -------
        objective_map : dict of [str, float]
          Mapping from objective names to objective values
        """
        named_parameters = { 'dai' : {} }
        for p, value in zip(self.parameters, parameter_values):
            kind = self.parameter_kind[p.name]
            if kind not in named_parameters:
                named_parameters[kind] = { p.name : value }
            else:
                named_parameters[kind][p.name] = value

        # If we debug then we dont want the directory to be deleted after use
        # From python 3.12 we can pass delete=False to TemporaryDirectory, but prior to that we need
        # to use mkdtemp.
        if self.debug:
            output_directory = tempfile.mkdtemp(dir=self.data_dir)
            return self._run(output_directory, named_parameters)

        with tempfile.TemporaryDirectory(dir=self.data_dir) as output_directory:
            return self._run(output_directory, named_parameters)

    def _run(self, output_directory, named_parameters):
        dai_file = self.file_generator(output_directory, named_parameters, tagged=True)['dai']
        sim_result = self.runner(dai_file, output_directory)
        if sim_result.returncode != 0:
            print(sim_result)
            return { self.objective_fn.name : np.nan }
        return self.objective_fn(output_directory)



'''Utility functions'''
from collections.abc import Sequence

__all__ = [
    'flatten'
    ]

def flatten(xs, f=None):
    '''Recursively flatten a sequence.

    Parameters
    ----------
    xs : Sequence
      The Sequence to flatten.

    f : Callable
      If not None apply f to all elements of the flattened list

    Returns
    -------
    Flattened representation of xs as a list

    Example
    -------
    >>> flatten([1, [2, 3, [4],], [[[5],6]]])
    [1, 2, 3, 4, 5, 6]

    >>> from collections.abc import Sequence
    ... class A:
    ...     def __init__(self, name):
    ...         self.name = name
    ...
    ... class B(Sequence):
    ...     def __init__(self, as_):
    ...         self.as_ = as_
    ...
    ...     @property
    ...     def name(self):
    ...         return flatten(self.as_, lambda a: a.name)
    ...
    ...     def __getitem__(self, index):
    ...         return self.As[index]
    ...
    ...     def __len__(self):
    ...         return len(self.As)
    ...
    >>> a1, a2, a3 = A('1'), A('2'), A('3')
    >>> print(a1.name, a2.name, a3.name)
    1 2 3

    >>> b1 = B([a1, a2]); b2 = B([a2, a3]); b3 = B([b1, b2])
    >>> print(b1.name, b2.name, b3.name)
    ['1', '2'] ['2', '3'] ['1', '2', '2', '3']
    '''
    if f is None:
        return _flatten_direct(xs)
    return _flatten_func(xs, f)

def _flatten_direct(xs):
    ys = []
    for x in xs:
        if isinstance(x, Sequence) and not isinstance(x, str):
            for y in _flatten_direct(x):
                ys.append(y)
        else:
            ys.append(x)
    return ys

def _flatten_func(xs, f):
    ys = []
    for x in xs:
        if isinstance(x, Sequence) and not isinstance(x, str):
            for y in _flatten_func(x, f):
                ys.append(y)
        else:
            ys.append(f(x))
    return ys



import os
from daisypy.optim import DefaultLogger

def test_default_logger(capsys, tmp_path):
    '''Test that DefaultLogger logs different messages to the expected locations'''
    expected_out = 'line 1 info'
    expected_err = '\n'.join([
        'line 1 warning',
        'line 1 error'
    ])
    expected_result = [
        'step,tag,value,p1,p2,p3',
        '1,"a",0.1,0,2,4'
    ]
    expected_result_path = tmp_path / 'result.csv'
    with DefaultLogger(tmp_path) as logger:
        logger.info('line 1 info')
        logger.warning('line 1 warning')
        logger.error('line 1 error')
        logger.result(step=1, tag='a', value=0.1, p1=0, p2=2, p3=4)

    assert os.path.exists(expected_result_path)
    with open(expected_result_path, 'r', encoding='utf-8') as in_file:
        lines = [line.strip() for line in in_file]
    assert lines == expected_result

    captured = capsys.readouterr()
    assert captured.out.strip() == expected_out
    assert captured.err.strip() == expected_err



from pathlib import Path
from daisypy.optim import (
    CategoricalParameter,
    DefaultLogger,
    DaisySequentialOptimizer,
    DlfDataExtractor,
    PyFileGenerator,
    DaiFileGenerator,
    MultiFileGenerator,
    DaisyRunner,
    DaisyOptimizationProblem,
    ScalarObjective,
    mse
)
from .markers import requires_daisy

@requires_daisy
def test_python_chemical_reaction(tmp_path):
    '''Test that we can optimize parameters of a python function in a full sim setup'''
    data_dir = Path(__file__).parent / 'test-data' / 'python-chemical-reaction'
    file_generator = MultiFileGenerator({
        'py' : PyFileGenerator('daisy-react.py', template_file_path=data_dir / 'daisy-react.py'),
        'dai' : DaiFileGenerator('run.dai', template_file_path=data_dir / 'test-pyreact.dai'),
    })

    parameters = {
        'py' : [ CategoricalParameter('param', [0.24, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6]) ],
        'dai' : [],
    }

    # Assume we are on linux and daisy is installed
    runner = DaisyRunner('daisy')

    target_file = data_dir / 'target.csv'
    data_extractor = DlfDataExtractor({"soil_NO3_profile.dlf" : "NO3"})
    objective = ScalarObjective("NO3", data_extractor, target_file, "NO3", mse)

    problem = DaisyOptimizationProblem(
        runner, file_generator, objective, parameters, tmp_path, debug=True
    )
    with DefaultLogger(tmp_path) as logger:
        optimizer = DaisySequentialOptimizer(problem, logger)
        result = optimizer.optimize()

    assert result['param']['best'] == 0.1



# pylint: disable=relative-beyond-top-level
import os
import tempfile
from daisypy.optim import (
    CategoricalParameter,
    DefaultLogger,
    DaisySequentialOptimizer,
)
from .mockup import MockProblem

class Objective:
    # pylint: disable=too-few-public-methods
    '''Negative sum of arguments'''
    def __init__(self, name):
        self.name = name

    def __call__(self, a, b, c):
        return - (a + b + c)

def test_sequential_optimizer(capsys):
    '''Test that sequential optimizer finds the optimmum and generates expected output'''
    expected_result_log = [
        'step,tag,metric_neg_sum,param_a,param_b,param_c',
        '1,"raw",-1.0,1.0,0.0,0.0',
        '1,"raw",-1.0,0.0,1.0,0.0',
        '1,"raw",-2.0,0.0,2.0,0.0',
        '1,"raw",-1.0,0.0,0.0,1.0',
        '1,"raw",-2.0,0.0,0.0,2.0',
        '1,"raw",-3.0,0.0,0.0,3.0',
        '2,"raw",-4.0,1.0,0.0,3.0',
        '2,"raw",-4.0,0.0,1.0,3.0',
        '2,"raw",-5.0,0.0,2.0,3.0',
        '3,"raw",-6.0,1.0,2.0,3.0',
    ]
    expected_out = '\n'.join([
        'Using at least 11 and at most 15 function evaluations',
        'Evaluating initial parameters',
        'Initial objective = 0',
        'Optimizing',
        'step=1,n_param_sets=6',
        'step=1,total_function_evaluations=7',
        'step=1,best_objective=-3.0',
        'step=1,Fixing c to 3',
        'step=2,n_param_sets=3',
        'step=2,total_function_evaluations=10',
        'step=2,best_objective=-5.0',
        'step=2,Fixing b to 2',
        'step=3,n_param_sets=1',
        'step=3,total_function_evaluations=11',
        'step=3,best_objective=-6.0',
        'step=3,Fixing a to 1',
    ])
    expected_err = ''
    parameters = [
        CategoricalParameter('a', [0,1]),
        CategoricalParameter('b', [0,1,2]),
        CategoricalParameter('c', [0,1,2,3]),
    ]

    problem = MockProblem(parameters, Objective("neg_sum"))
    with tempfile.TemporaryDirectory() as out_dir:
        with DefaultLogger(out_dir) as logger:
            optimizer = DaisySequentialOptimizer(problem, logger)
            result = optimizer.optimize()
        with open(os.path.join(out_dir, 'result.csv'), 'r', encoding='utf-8') as in_file:
            for expected, row in zip(expected_result_log, in_file, strict=True):
                assert expected == row.strip()

    captured = capsys.readouterr()
    assert captured.out.strip() == expected_out
    assert captured.err.strip() == expected_err

    assert result['a']['best'] == 1
    assert result['b']['best'] == 2
    assert result['c']['best'] == 3



from collections.abc import Sequence
from .util import flatten

class MultiObjective(Sequence):
    '''Objective that computes several objectives.'''

    def __init__(self, name, objective_fns):
        '''
        Parameters
        ----------
        name : str
          Name of objective

        objective_fns : list of Callable[[str], float]
          List of objective functions. The objective function is passed the
          path to a daisy output directory and is expected to return a scalar
        '''
        self.name = name
        self.objective_fns = objective_fns

    def __call__(self, daisy_output_directory):
        '''Compute the objectives

        Parameters
        ----------
        daisy_output_directory : str
          Path to daisy ouput directory that is used when calling the objective functions

        Returns
        -------
        objective_map : dict of [str, float]
          Mapping from objective names to objective values
        '''
        return { k:v for f in self.objective_fns for k,v in f(daisy_output_directory).items() }

    def __getitem__(self, index):
        return self.objective_fns[index]

    def __len__(self):
        return len(self.objective_fns)

    @property
    def variable_name(self):
        '''Names of variables used in aggregated objectives

        Returns
        -------
        list of str
        '''
        return flatten(self.objective_fns, lambda x : x.variable_name)

    @property
    def target(self):
        '''Targets used on aggregated objectives

        Returns
        -------
        list of pandas.DataFrame
        '''
        return flatten(self.objective_fns, lambda x : x.target)

    @property
    def log_name(self):
        '''Names of Daisy log files used in aggregated objectives

        Returns
        -------
        list of str
        '''
        return flatten(self.objective_fns, lambda x : x.log_name)



# pylint: disable=relative-beyond-top-level
import tempfile
from pytest import approx
from daisypy.optim import (
    DefaultLogger,
    DaisyCMAOptimizer,
)
from .mockup import MockProblem
from .test_objectives import beale_function

def test_cma_optimizer():
    '''Test that CMA can optimize the Beale function'''
    problem = MockProblem(beale_function.parameters, beale_function)
    out_dir = 'out/cma'
    with tempfile.TemporaryDirectory() as out_dir:
        with DefaultLogger(out_dir) as logger:
            optimizer = DaisyCMAOptimizer(problem, logger, cma_options = { "maxfevals" : 500 })
            result = optimizer.optimize()
    for k,v in result.items():
        assert v['mean_transformed'] == approx(beale_function.amin[k])



# pylint: disable=too-few-public-methods
class ContinuousParameter:
    """Wrapper for continuous Daisy parameters"""
    def __init__(self, name, initial_value, valid_range):
        """
        Parameters
        ----------
        name : str
          Name of parameter

        initial_value : float
          Initial value of parameter to use for starting the optimization

        valid_range : (float, float)
          Lowest and highest valid values for the parameter
        """
        self.type = "Continuous"
        self.name = name
        self.initial_value = initial_value
        self.valid_range = valid_range


class CategoricalParameter:
    """Wrapper for categorical parameters"""
    def __init__(self, name, values, initial_value_idx=0):
        """
        Parameters
        ----------
        name : str
          Name of parameter

        values : list of values
          Possible parameter values

        initial_value_idx : int
          Index of initial parameter value
        """
        self.type = "Categorical"
        self.name = name
        self.values = values
        self.initial_value_idx = initial_value_idx



import os
import subprocess

class DaisyRunner:
    """Class that knows how to run run daisy"""

    def __init__(self, daisy_bin, daisy_home=None):
        """
        Parameters
        ----------
        daisy_bin : str
          Path to daisy binary

        daisy_home : str
          Path to daisy home directory containing lib/ and sample/
          If not None set DAISYHOME environment variable to daisy_home. Otherwise dont set DAISYHOME
        """
        self.daisy_bin = daisy_bin
        if daisy_home is not None:
            os.environ.update('DAISYHOME', daisy_home)

    def __call__(self, dai_file, output_directory):
        """Run daisy

        Parameters
        ----------
        dai_file : str
          Path to dai file to run

        output_directory : str
          Path to output directory

        Returns
        -------
        subprocess.CompletedProcess
        """
        args = [
            self.daisy_bin,
            "-q",
            "-d", output_directory,
            dai_file
        ]
        return subprocess.run(args, check=False)


    def serialize(self):
        """Serialize this DaisyRunner object

        Returns
        -------
        dict of (parameter name, parameter value) pairs
        """
        return {
            'daisy_bin' : self.daisy_bin,
            'daisy_home' : os.environ.get('DAISYHOME', None)
        }

    @staticmethod
    def unzerialize(dict_repr):
        """Unserialize a DaisyRunner

        Parameters
        ----------
        dict_repr : dict of (str, str)
          Must contain
            daisy_bin : path to daisy binary
            daisy_home : path to daisy home
        """
        return DaisyRunner(dict_repr['daisy_bin'], dict_repr['daisy_home'])



# pylint: disable=missing-function-docstring,R0801
from pathlib import Path
import pandas as pd
import pytest
from daisypy.optim import DlfDataExtractor, DlfSum
from daisypy.optim.dlf_data_extraction import DlfSingleton

EXPECTED_NO3 = pd.Series([
    0.000114649, 8.02047e-05, 5.61092e-05, 3.92532e-05, 2.74616e-05, 1.92128e-05, 1.34425e-05,
    9.40583e-06, 6.58209e-06, 4.60684e-06, 3.22518e-06, 2.25878e-06, 1.58289e-06, 1.11026e-06,
    7.79861e-07, 5.48966e-07, 3.87692e-07, 2.75131e-07, 1.96658e-07, 1.42039e-07, 1.04112e-07,
    7.78724e-08, 5.98114e-08, 4.74821e-08
])

EXPECTED_CO2 = pd.Series([
    -76.3192, -76.3185, -76.3178, -76.3171, -76.3164, -76.3157, -76.315	, -76.3143, -76.3136,
    -76.3129, -76.3122, -76.3115, -76.3108, -76.3101, -76.3094, -76.3087, -76.3079, -76.3072,
    -76.3065, -76.3058, -76.3051, -76.3044, -76.3037, -76.303
])

def test_single_var_no_post_processor():
    data_dir = Path(__file__).parent / 'test-data' / 'dlfs'
    extractor = DlfDataExtractor({'soil_NO3_profile.dlf' : 'NO3'})
    extracted = extractor(data_dir)
    assert set(extracted.columns) == { 'time' , 'value' }
    assert (EXPECTED_NO3 == extracted['value']).all()

def test_single_var_singleton_post_processor():
    data_dir = Path(__file__).parent / 'test-data' / 'dlfs'
    extractor = DlfDataExtractor({'soil_NO3_profile.dlf' : 'NO3'},  DlfSingleton())
    extracted = extractor(data_dir)
    assert set(extracted.columns) == { 'time' , 'value' }
    assert (EXPECTED_NO3 == extracted['value']).all()

def test_multiple_vars_no_post_processor():
    with pytest.raises(ValueError) as excinfo:
        _ = DlfDataExtractor({'soil_NO3_profile.dlf' : ['NO3', 'CO2']})
    assert "`logs_and_variables` must contain exactly one key" in str(excinfo.value)

def test_bad_post_processor():
    data_dir = Path(__file__).parent / 'test-data' / 'dlfs'
    extractor = DlfDataExtractor({'soil_NO3_profile.dlf' : 'NO3'},  lambda dfs: dfs[0])
    # Post processor does not ensure correct naming of columns, should raise an exception
    with pytest.raises(RuntimeError) as excinfo:
        _ = extractor(data_dir)
    assert "must return a DataFrame with columns 'time' and 'value'" in str(excinfo.value)

def test_multiple_vars_sum_post_processor():
    data_dir = Path(__file__).parent / 'test-data' / 'dlfs'
    extractor = DlfDataExtractor({'soil_NO3_profile.dlf' : ['NO3', 'CO2']}, DlfSum())
    extracted = extractor(data_dir)
    assert set(extracted.columns) == { 'time' , 'value' }
    assert (EXPECTED_NO3 + EXPECTED_CO2 == extracted['value']).all()

def test_bad_logs_and_variables():
    with pytest.raises(ValueError) as excinfo:
        # Pass what should be a dict as two separate params
        _ = DlfDataExtractor('soil_NO3_profile.dlf', 'NO3')
    assert '`logs_and_variables` must be a Mapping' in str(excinfo.value)



import warnings
import random
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib import animation
from matplotlib.colors import Normalize
import matplotlib as mpl

__all__ = [
    'plot_convergence',
    'plot_result',
    'plot_result_1d',
    'plot_result_2d',
    'plot_result_3d',
    'plot_result_nd',
    'animate_result',
    'animate_result_1d',
    'animate_result_2d',
    'animate_result_nd',
]

def plot_convergence(df, step_var='step', f_var='objective_value', log_transform=True):
    '''Make a convergence plot of objective values.

    Parameters
    ----------
    df : pandas.DataFrame OR str
      If str assume it is a path to a csv file with optimization results

    step_var : str
      Name of step variable

    f_var : str
      Name of objective variable

    log_transform : bool
      If True apply a log transform to the objective value. The objective is shifted to be positive
      before log transforming.

    Returns
    -------
    (matplotlib.Figure, matplotlib.Axes)
    '''
    if not isinstance(df, pd.DataFrame):
        df = pd.read_csv(df)
    step = df[step_var]
    f = df[f_var]
    if log_transform:
        if np.min(f) <= 0:
            f = 1e-8 + f - np.min(f)
        f = np.log(f)
    fig, ax = plt.subplots()
    even = step % 2 == 0
    odd = ~even
    ax.scatter(step[even], f[even], marker='.', c='black')
    ax.scatter(step[odd], f[odd], marker='+', c='black')
    ax.set_xlabel(step_var)
    ax.set_ylabel('log(f)')
    return fig, ax

def plot_result(df, step_var='step', f_var='objective_value'):
    '''Make a scatter plot of parameters and objective values at each optimization step. If there is
    more than 3 parameters, each parameter is plotted in its own subplot. If there is more than 20
    parameters, each set of 20 parameters is plotted in a separate figure.

    Parameters
    ----------
    df : pandas.DataFrame OR str
      If str assume it is a path to a csv file with optimization results

    step_var : str
      Name of step variable

    f_var : str
      Name of objective variable

    Returns
    -------
    (matplotlib.Figure, matplotlib.Axes) or list of (matplotlib.Figure, matplotlib.Axes)

    See also
    --------
    plot_result_1d, plot_result_2d, plot_result_3d, plot_result_nd
    '''
    if not isinstance(df, pd.DataFrame):
        df = pd.read_csv(df)
    params = df.columns[~df.columns.isin({step_var, f_var})]
    match len(params):
        case 1: return plot_result_1d(df, params[0], step_var, f_var)
        case 2: return plot_result_2d(df, params[0], params[1], step_var, f_var)
        case 3: return plot_result_3d(df, params[0], params[1], params[2], f_var)
        case _: return plot_result_nd(df, params, step_var, f_var)

def plot_result_1d(df, var, step_var='step', f_var='objective_value'):
    '''Make a scatter plot of a single parameter and objective value at each optimization step.
    Parameters
    ----------
    df : pandas.DataFrame OR str
      If str assume it is a path to a csv file with optimization results

    var : str
      Name of parameter variable

    step_var : str
      Name of step variable

    f_var : str
      Name of objective variable

    Returns
    -------
    (matplotlib.Figure, matplotlib.Axes)

    See also
    --------
    plot_result, plot_result_2d, plot_result_3d, plot_result_nd
    '''
    if not isinstance(df, pd.DataFrame):
        df = pd.read_csv(df)
    fig, ax = plt.subplots()
    x = df[step_var]
    y = df[var]
    f = df[f_var]
    transform, transform_text = _find_cmap_transform(f)
    c = transform(f)
    cm = ax.scatter(x, y, c=c)
    ax.set_xlabel(step_var)
    ax.set_ylabel(var)
    ax.set_title('Parameters at each step')
    fig.colorbar(cm, label=transform_text)
    fig.tight_layout()
    return fig, ax

def plot_result_2d(df, x_var, y_var, step_var='step', f_var='objective_value'):
    '''Make a scatter plot of two parameters and objective value at each optimization step.

    Parameters
    ----------
    df : pandas.DataFrame OR str
      If str assume it is a path to a csv file with optimization results

    x_var : str
      Name of parameter variable on x-axis

    y_var : str
      Name of parameter variable on y-axis

    step_var : str
      Name of step variable

    f_var : str
      Name of objective variable

    Returns
    -------
    (matplotlib.Figure, matplotlib.Axes)

    See also
    --------
    plot_result, plot_result_1d, plot_result_3d, plot_result_nd
    '''
    if not isinstance(df, pd.DataFrame):
        df = pd.read_csv(df)
    fig, ax = plt.subplots()
    transform, transform_text = _find_cmap_transform(df[f_var])
    c = transform(df[f_var])
    norm = Normalize(vmin=np.min(c), vmax=np.max(c))
    grouped = pd.DataFrame({
        'x': df[x_var], 'y': df[y_var], 'c': c, 'step': df[step_var]
    }).groupby('step')
    for step, group in grouped:
        cm = ax.scatter(group['x'], group['y'], c=group['c'], marker=f'${step}$', norm=norm)
    ax.set_xlabel(x_var)
    ax.set_ylabel(y_var)
    ax.set_title(f'Parameters at each step for {len(grouped)} steps')
    fig.colorbar(cm, label=transform_text)
    fig.tight_layout()
    return fig, ax

def plot_result_3d(df, x_var, y_var, z_var, f_var):
    '''Make a 3D scatter plot of three parameters and objective values.

    Parameters
    ----------
    df : pandas.DataFrame OR str
      If str assume it is a path to a csv file with optimization results

    x_var : str
      Name of parameter variable on x-axis

    y_var : str
      Name of parameter variable on y-axis

    z_var : str
      Name of parameter variable on z-axis

    f_var : str
      Name of objective variable

    Returns
    -------
    (matplotlib.Figure, matplotlib.Axes)

    See also
    --------
    plot_result, plot_result_1d, plot_result_2d, plot_result_nd
    '''
    if not isinstance(df, pd.DataFrame):
        df = pd.read_csv(df)
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    x = df[x_var]
    y = df[y_var]
    z = df[z_var]
    f = df[f_var]
    transform, _ = _find_cmap_transform(f)
    c = transform(f)
    ax.scatter(x, y, z, c=c)
    return fig, ax

def plot_result_nd(df, params, step_var='step', f_var='objective_value', max_plots_in_figure=20):
    '''Plot each parameter in a separate subplot, showing the parameter value and objective value at
    each step. If there is more than 20 parameters, each set of 20 parameters is plotted in a
    separate figure.

    Parameters
    ----------
    df : pandas.DataFrame OR str
      If str assume it is a path to a csv file with optimization results

    params : list of str
      Names of parameter variables to plot

    step_var : str
      Name of step variable

    f_var : str
      Name of objective variable

    max_plots_in_figure : int
      Maximum number of parameters to plot per figure

    Returns
    -------
    (matplotlib.Figure, matplotlib.Axes) or list of (matplotlib.Figure, matplotlib.Axes)

    See also
    --------
    plot_result, plot_result_1d, plot_result_2d, plot_result_3d
    '''
    # pylint: disable=too-many-locals
    if not isinstance(df, pd.DataFrame):
        df = pd.read_csv(df)
    n = len(params)
    if n > max_plots_in_figure:
        return [
            plot_result_nd(df, params[i:min(n, i+max_plots_in_figure)], step_var, f_var)
            for i in range(n, step=max_plots_in_figure)
        ]

    x = df[step_var]
    c, transform_text = _cmap_transform(df[f_var])

    rows = int(np.sqrt(n))
    cols = n // rows
    if rows * cols < n:
        cols += 1
    fig, axs = plt.subplots(rows, cols, squeeze=False)
    row, col = 0, 0
    for var in params:
        ax = axs[row,col]
        cm = ax.scatter(x, df[var], c=c)
        ax.set_xlabel(step_var)
        ax.set_ylabel(var)
        col = (col + 1) % cols
        if col == 0:
            row = row + 1
    fig.colorbar(cm, ax=axs, label=transform_text)
    return fig, ax

def animate_result(df, step_var='step', f_var='objective_value'):
    '''Make an animation showing parameters at each optimization step. If there are more than two
    parameters, two are selected at random and animated.

    Parameters
    ----------
    df : pandas.DataFrame OR str
      If str assume it is a path to a csv file with optimization results

    step_var : str
      Name of step variable

    f_var : str
      Name of objective variable

    Returns
    -------
    matplotlib.animation.AnimationFigure

    See also
    --------
    animate_result_1d, animate_result_2d, animate_result_nd
    '''
    if not isinstance(df, pd.DataFrame):
        df = pd.read_csv(df)
    params = df.columns[~df.columns.isin({step_var, f_var})]

    match len(params):
        case 1: return animate_result_1d(df, params[0], step_var, f_var)
        case 2: return animate_result_2d(df, params[0], params[1], step_var, f_var)
        case _: return animate_result_nd(df, params, step_var, f_var)

def animate_result_1d(df, var, step_var='step', f_var='objective_value'):
    '''Make an animation showing a single parameter at each optimization step.

    Parameters
    ----------
    df : pandas.DataFrame OR str
      If str assume it is a path to a csv file with optimization results

    var : str
      Name of parameter variable

    step_var : str
      Name of step variable

    f_var : str
      Name of objective variable

    Returns
    -------
    matplotlib.animation.AnimationFigure

    See also
    --------
    animate_result, animate_result_2d, animate_result_nd
    '''
    # pylint: disable=too-many-locals
    if not isinstance(df, pd.DataFrame):
        df = pd.read_csv(df)
    # TODO: Add boundary lines showing parameter extents
    cmap = mpl.colormaps[mpl.rcParams['image.cmap']]

    # Find x and y limits that cover full data set
    xlim = _compute_limits(df[step_var])
    ylim = _compute_limits(df[var])

    # Find a suitable transformation of the objective for coloring
    transform, transform_text = _find_cmap_transform(df[f_var])
    c = transform(df[f_var])
    norm = Normalize(vmin=np.min(c), vmax=np.max(c))

    # Prepare the data so we can just index it
    xs, ys, cs = [], [], []
    for step, group in df.groupby(step_var):
        xs.append([step]*len(group))
        ys.append(group[var])
        cs.append(cmap(norm(transform(group[f_var]))))

    fig, ax = plt.subplots(1)
    ax.set_xlabel(step_var)
    ax.set_ylabel(var)
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    # Make a colormap that has correct labeling and values
    dummy_plot = ax.scatter([], [], c=[], cmap=cmap, norm=norm)
    fig.colorbar(dummy_plot, label=transform_text)

    plot = ax.scatter(xs[0], ys[0], c=cs[0])
    def update(frame):
        data = np.stack([xs[frame], ys[frame]]).T
        plot.set_offsets(data)
        plot.set_color(cs[frame])
    n_frames = len(xs)
    interval = max(100, 2000//n_frames)
    return animation.FuncAnimation(fig=fig, func=update, frames=n_frames, interval=interval)


def animate_result_2d(df, x_var, y_var, step_var='step', f_var='objective_value'):
    '''Make an animation showing two parameters at each optimization step.

    Parameters
    ----------
    df : pandas.DataFrame OR str
      If str assume it is a path to a csv file with optimization results

    x_var : str
      Name of parameter variable on x-axis

    y_var : str
      Name of parameter variable on y-axis

    step_var : str
      Name of step variable

    f_var : str
      Name of objective variable

    Returns
    -------
    matplotlib.animation.AnimationFigure

    See also
    --------
    animate_result, animate_result_1d, animate_result_nd
    '''
    # pylint: disable=too-many-locals
    if not isinstance(df, pd.DataFrame):
        df = pd.read_csv(df)
    cmap = mpl.colormaps[mpl.rcParams['image.cmap']]

    # Find x and y limits that cover full data set
    xlim = _compute_limits(df[x_var])
    ylim = _compute_limits(df[y_var])

    # Find a suitable transformation of the objective for coloring
    transform, transform_text = _find_cmap_transform(df[f_var])
    c = transform(df[f_var])
    norm = Normalize(vmin=np.min(c), vmax=np.max(c))

    # Prepare the data so we can just index it
    steps, xs, ys, cs = [], [], [], []
    for step, group in df.groupby(step_var):
        steps.append(step)
        xs.append(group[x_var])
        ys.append(group[y_var])
        cs.append(cmap(norm(transform(group[f_var]))))

    fig, ax = plt.subplots(1)
    ax.set_xlabel(x_var)
    ax.set_ylabel(y_var)
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    ax.set_title(f'Step {steps[0]}')

    # Make a colormap that has correct labeling and values
    dummy_plot = ax.scatter([], [], c=[], cmap=cmap, norm=norm)
    fig.colorbar(dummy_plot, label=transform_text)

    plot = ax.scatter(xs[0], ys[0], c=cs[0])
    def update(frame):
        data = np.stack([xs[frame], ys[frame]]).T
        plot.set_offsets(data)
        plot.set_color(cs[frame])
        ax.set_title(f'Step {steps[frame]}')
    n_frames = len(xs)
    interval = max(100, 2000//n_frames)
    return animation.FuncAnimation(fig=fig, func=update, frames=n_frames, interval=interval)

def animate_result_nd(df, params, step_var='step', f_var='objective_value'):
    '''Make an animation of two parameters selected at random from a list of possible parameters

    Parameters
    ----------
    df : pandas.DataFrame OR str
      If str assume it is a path to a csv file with optimization results

    params : list of str
      Names of parameter variables

    step_var : str
      Name of step variable

    f_var : str
      Name of objective variable

    Returns
    -------
    matplotlib.animation.AnimationFigure

    See also
    --------
    animate_result, animate_result_1d, animate_result_nd
    '''

    warnings.warn("Can only animate 2 parameters at a time. Selecting two at random")
    params_list = list(params)
    random.shuffle(params_list)
    x_var = params_list[0]
    y_var = params_list[1]
    return animate_result_2d(df, x_var, y_var, step_var, f_var)

## Utilities
def _compute_limits(x):
    lim = [np.min(x), np.max(x)]
    w = (lim[1] - lim[0]) * 1.1
    c = (lim[0] + lim[1]) / 2
    return (c - w/2, c + w/2)

def _shift_and_scale(x):
    shifted = x - np.min(x)
    return shifted / np.max(shifted)

def _linearity(f):
    f = _shift_and_scale(f) # Standardize to [0,1]
    x = np.arange(len(f))
    n = x[-1]
    target = x / n * (f[n] - f[0]) + f[0] # Straight line connecting end points
    return np.mean((f - target))


def _cmap_transform(f):
    transform, transform_text = _find_cmap_transform(f)
    return transform(f), transform_text

def _find_cmap_transform(f):
    # Try to find a transformation of f such that it looks nice when used with linear color maps
    lin_score = _linearity(f)
    if lin_score < -0.2:
        return lambda x : np.log(1e-8 + x - np.min(f)), "log(f)"
    if lin_score < -0.1:
        return lambda x : np.sqrt(x - np.min(f)), "$\\sqrt{f}$"
    if lin_score > 0.2:
        return np.exp, "exp(f)"
    if lin_score > 0.1:
        return lambda x : x**2, "$f^2$"
    return lambda x : x, "f"



from pathlib import Path
from daisypy.optim import DaisyRunner
from .markers import requires_daisy

EXPECTED = "Hello from Daisy"

@requires_daisy
def test_runner(tmp_path):
    '''Test that DaisyRunner can run Daisy and generate the expected daisy.log'''
    # Assume we are on linux and daisy is installed
    runner = DaisyRunner('daisy')
    dai_path = Path(__file__).parent / 'hello.dai'
    result = runner(str(dai_path), str(tmp_path))
    assert result.returncode == 0

    daisy_log = tmp_path / 'daisy.log'
    with daisy_log.open(encoding='utf-8') as f:
        lines = list(f)
    assert len(lines) >= 2
    assert lines[-2].strip() == EXPECTED



# pylint: disable=missing-function-docstring,R0801
from pathlib import Path
import pytest
from daisypy.optim.dai_file_generator import DaiFileGenerator

EXPECTED = """(deffunction f Python
  "Call Python function." (module "testing") (name "linear") (domain []) (range []))
(defprogram print_it write
  "Write specific value"
  (declare v1 Number [] "V1")
  (declare v2 Number [] "V2")
  (v1 (apply f 0 []))
  (v2 (apply f 5 []))
  (what "f(0) = ${v1}, f(5) = ${v2}"))
(run print_it)"""

PARAMS = { 'x1' : 0, 'x2' : 5 }

def test_tagged(tmp_path):
    '''Test that generated dai file is as expected'''
    template_path = Path(__file__).parent / 'templates' / 'template.dai'
    generator = DaiFileGenerator('linear.dai', template_file_path=template_path)
    file_path = Path(generator(tmp_path, {'dai' : PARAMS})['dai'])
    assert file_path.read_text(encoding='utf-8') == EXPECTED

def test_not_tagged(tmp_path):
    template_path = Path(__file__).parent / 'templates' / 'template.dai'
    generator = DaiFileGenerator('linear.dai', template_file_path=template_path)
    file_path = Path(generator(tmp_path, PARAMS, tagged=False))
    assert file_path.read_text(encoding='utf-8') == EXPECTED

def test_no_params(tmp_path):
    template = '(defprogram print_it write\n  (what "${{v1}}"))'
    expected = '(defprogram print_it write\n  (what "${v1}"))'
    generator = DaiFileGenerator('linear.dai', template_text=template)
    file_path = Path(generator(tmp_path, {}, tagged=False))
    assert file_path.read_text(encoding='utf-8') == expected
    file_path = Path(generator(tmp_path, {'dai' : {}})['dai'])
    assert file_path.read_text(encoding='utf-8') == expected

SPAWN_PARALLEL = """(defprogram p1 spawn (parallel 10))"""
SPAWN_SEQUENTIAL = """(defprogram p1 spawn\n  (parallel 1))"""

def test_spawn_is_made_sequential():
    with pytest.warns(UserWarning, match="spawn forced to 1"):
        generator = DaiFileGenerator('dummy', template_text=SPAWN_PARALLEL)
    assert generator.template_text == SPAWN_SEQUENTIAL



# pylint: disable=R0801
import multiprocessing
from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass
from ax.api.client import Client
from .ax import daisy_param_to_ax_param
from .multi_objective import MultiObjective

@dataclass
class AxResult:
    '''Class holding Ax optimization result'''
    parameters : dict
    metrics : dict

class DaisyAxOptimizer:
    # pylint: disable=too-few-public-methods,too-many-locals
    """Daisy optimizer using Ax. Can do scalar and multi objective optimization"""
    def __init__(self, problem, logger, options=None, number_of_processes=None):
        """
        Parameters
        ----------
        problem : DaisyProblem

        options : dict
        """
        self.problem = problem
        self.logger = logger
        if number_of_processes is None:
            self.number_of_processes = multiprocessing.cpu_count()
        else:
            self.number_of_processes = number_of_processes

        if options is None:
            options = {}
        options.setdefault('max_trials', 10)
        options.setdefault('max_trials_iteration', 3)
        self.options = options

        ax_parameters = [ daisy_param_to_ax_param(p) for p in self.problem.parameters ]

        self.client = Client()
        self.client.configure_experiment(parameters=ax_parameters)

        # TODO: Assumes we minimize
        self.multi_objective = isinstance(problem.objective_fn, MultiObjective)
        if self.multi_objective:
            objective_str = ','.join([f'-{f.name}' for f in problem.objective_fn.objective_fns])
        else:
            objective_str = f'-{problem.objective_fn.name}'
        self.client.configure_optimization(objective=objective_str)


    def optimize(self):
        '''Run the optimizer and return the result. The result is a single AxResult when doing
        scalar optimization and a list of AxResult when doing multi optimization

        Returns
        -------
        AxResult OR list of AxResult
        '''
        # TODO: Log parameter distributions
        num_trials = 0
        max_trials = self.options['max_trials']
        max_trials_iteration = self.options['max_trials_iteration']
        with ProcessPoolExecutor(self.number_of_processes) as executor:
            while num_trials < self.options['max_trials']:
                max_trials_this_iteration = min(max_trials_iteration, max_trials - num_trials)
                trials = self.client.get_next_trials(max_trials=max_trials_this_iteration)
                trial_indices = []
                parameter_sets = []
                named_parameter_sets = []
                for trial_index, sampled_parameters in trials.items():
                    trial_indices.append(trial_index)
                    named_params = {
                        p.name : sampled_parameters[p.name] for p in self.problem.parameters
                    }
                    named_parameter_sets.append(named_params)
                    params = [sampled_parameters[p.name] for p in self.problem.parameters]
                    parameter_sets.append(params)

                # Run simulations in parallel
                for i, result in enumerate(executor.map(self.problem, parameter_sets)):
                    log = { 'trial' : trial_indices[i] }
                    for name, value in named_parameter_sets[i].items():
                        log[f'param_{name}'] = value
                    for name, value in result.items():
                        log[f'metric_{name}'] = value
                    self.logger.result(**log)
                    self.client.complete_trial(trial_index=trial_indices[i], raw_data=result)
                num_trials += len(trials)

        if self.multi_objective:
            # Handle multi objective result
            result = [
                AxResult(parameters, metrics)
                for parameters, metrics, _, _ in self.client.get_pareto_frontier()
            ]
        else:
            # Handle scalar objective result
            parameters, metrics, _, _ = self.client.get_best_parameterization()
            result = AxResult(parameters, metrics)
        return result



"""Scalar loss functions.
Signature is (actual : numpy.ndarray, target : numpy.ndarray) -> float
"""

def mse(actual, target):
    """Mean squared error"""
    return ((actual - target)**2).mean()

def mae(actual, target):
    """Mean absolute error"""
    return abs(actual - target).mean()

available_loss_fns = {
    "mse" : mse,
    "mae" : mae
}



import numpy as np
from daisypy.optim import DaisyOptimizationProblem, ContinuousParameter, MultiObjective
from .mockup import (MockRunner, MockFileGenerator, MockObjective)


def test_runner_succeds(tmp_path):
    '''Test that the return value is as expected when the runner succeds'''
    file_generator = MockFileGenerator({'dai' : ''})
    runner = MockRunner()
    parameters = { 'dai' : [ContinuousParameter('p', 0, (-1, 1))] }
    out_dir = tmp_path
    objective = MockObjective('mock', 123)

    problem = DaisyOptimizationProblem(
        runner, file_generator, objective, parameters, out_dir
    )
    result = problem([-1])
    assert result['mock'] == objective.value


def test_runner_fails(tmp_path):
    '''Test that the return value is nan when the runner fails'''
    file_generator = MockFileGenerator({'dai' : ''})
    runner = MockRunner(returncode=1)
    parameters = { 'dai' : [ContinuousParameter('p', 0, (-1, 1))] }
    out_dir = tmp_path
    objective = MockObjective('mock', 123)

    problem = DaisyOptimizationProblem(
        runner, file_generator, objective, parameters, out_dir
    )
    result = problem([-1])
    assert np.isnan(result['mock'])

def test_multi_objective(tmp_path):
    '''Test that the return value is as expected for multiple objectives'''
    file_generator = MockFileGenerator({'dai' : ''})
    runner = MockRunner()
    parameters = { 'dai' : [ContinuousParameter('p', 0, (-1, 1))] }
    out_dir = tmp_path
    objectives = [ MockObjective(f'mock-{i}', i*123) for i in range(3) ]
    objective = MultiObjective('multi', objectives)

    problem = DaisyOptimizationProblem(
        runner, file_generator, objective, parameters, out_dir
    )
    result = problem([0])
    for obj in objectives:
        assert result[obj.name] == obj.value
